# Generated from OliviaCSParser.g4 by ANTLR 4.13.0
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,42,103,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,1,0,3,0,16,8,0,1,0,1,0,1,0,1,0,5,0,22,8,0,10,0,12,0,25,9,0,1,0,
        1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
        1,1,4,1,45,8,1,11,1,12,1,46,1,1,1,1,1,2,1,2,1,2,1,2,1,2,1,3,1,3,
        1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,
        1,3,1,3,1,3,1,3,4,3,78,8,3,11,3,12,3,79,1,3,1,3,1,4,1,4,1,4,1,4,
        1,4,1,4,1,4,1,4,1,5,1,5,1,6,1,6,5,6,96,8,6,10,6,12,6,99,9,6,1,6,
        1,6,1,6,0,0,7,0,2,4,6,8,10,12,0,3,1,0,24,26,1,0,19,21,1,0,7,9,103,
        0,15,1,0,0,0,2,28,1,0,0,0,4,50,1,0,0,0,6,55,1,0,0,0,8,83,1,0,0,0,
        10,91,1,0,0,0,12,93,1,0,0,0,14,16,5,15,0,0,15,14,1,0,0,0,15,16,1,
        0,0,0,16,23,1,0,0,0,17,22,3,12,6,0,18,22,3,2,1,0,19,22,3,6,3,0,20,
        22,5,15,0,0,21,17,1,0,0,0,21,18,1,0,0,0,21,19,1,0,0,0,21,20,1,0,
        0,0,22,25,1,0,0,0,23,21,1,0,0,0,23,24,1,0,0,0,24,26,1,0,0,0,25,23,
        1,0,0,0,26,27,5,0,0,1,27,1,1,0,0,0,28,29,5,1,0,0,29,30,5,18,0,0,
        30,31,5,23,0,0,31,32,5,2,0,0,32,33,7,0,0,0,33,34,5,28,0,0,34,35,
        5,3,0,0,35,36,5,29,0,0,36,37,5,30,0,0,37,38,5,31,0,0,38,39,5,36,
        0,0,39,40,5,37,0,0,40,41,5,37,0,0,41,42,5,37,0,0,42,44,5,39,0,0,
        43,45,3,4,2,0,44,43,1,0,0,0,45,46,1,0,0,0,46,44,1,0,0,0,46,47,1,
        0,0,0,47,48,1,0,0,0,48,49,5,4,0,0,49,3,1,0,0,0,50,51,5,13,0,0,51,
        52,5,13,0,0,52,53,5,7,0,0,53,54,5,15,0,0,54,5,1,0,0,0,55,56,5,1,
        0,0,56,57,7,1,0,0,57,58,5,23,0,0,58,59,5,2,0,0,59,60,7,0,0,0,60,
        61,5,28,0,0,61,62,5,3,0,0,62,63,5,29,0,0,63,64,5,30,0,0,64,65,5,
        31,0,0,65,66,5,32,0,0,66,67,5,33,0,0,67,68,5,34,0,0,68,69,5,36,0,
        0,69,70,5,37,0,0,70,71,5,37,0,0,71,72,5,37,0,0,72,73,5,37,0,0,73,
        74,5,37,0,0,74,75,5,37,0,0,75,77,5,39,0,0,76,78,3,8,4,0,77,76,1,
        0,0,0,78,79,1,0,0,0,79,77,1,0,0,0,79,80,1,0,0,0,80,81,1,0,0,0,81,
        82,5,4,0,0,82,7,1,0,0,0,83,84,5,13,0,0,84,85,5,13,0,0,85,86,5,7,
        0,0,86,87,5,13,0,0,87,88,3,10,5,0,88,89,3,10,5,0,89,90,5,15,0,0,
        90,9,1,0,0,0,91,92,7,2,0,0,92,11,1,0,0,0,93,97,5,10,0,0,94,96,5,
        40,0,0,95,94,1,0,0,0,96,99,1,0,0,0,97,95,1,0,0,0,97,98,1,0,0,0,98,
        100,1,0,0,0,99,97,1,0,0,0,100,101,5,42,0,0,101,13,1,0,0,0,6,15,21,
        23,46,79,97
    ]

class OliviaCSParser ( Parser ):

    grammarFileName = "OliviaCSParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'TYPEDEF'", "'SEPARATOR'", "'FORMAT\\n'", 
                     "'UNFORMAT'", "'EOF'", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "'REMARK'", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "'SEQUENCE'", "'ASS_TBL_H2O'", "'ASS_TBL_TRO'", 
                     "'ASS_TBL_D2O'", "<INVALID>", "<INVALID>", "'TAB'", 
                     "'COMMA'", "'SPACE'", "<INVALID>", "<INVALID>", "'CHAIN'", 
                     "'RESNAME'", "'SEQNUM'", "'ATOMNAME'", "'SHIFT'", "'STDDEV'" ]

    symbolicNames = [ "<INVALID>", "Typedef", "Separator", "Format", "Unformat", 
                      "Eof", "Null_string", "Integer", "Float", "Real", 
                      "COMMENT", "SHARP_COMMENT", "EXCLM_COMMENT", "Simple_name", 
                      "SPACE", "RETURN", "SECTION_COMMENT", "LINE_COMMENT", 
                      "Sequence", "Ass_tbl_h2o", "Ass_tbl_tro", "Ass_tbl_d2o", 
                      "SPACE_TD", "RETURN_TD", "Tab", "Comma", "Space", 
                      "SPACE_SE", "RETURN_SE", "Chain", "Resname", "Seqnum", 
                      "Atomname", "Shift", "Stddev", "SPACE_FO", "RETURN_FO", 
                      "Printf_string", "SPACE_PF", "RETURN_PF", "Any_name", 
                      "SPACE_CM", "RETURN_CM" ]

    RULE_olivia_cs = 0
    RULE_sequence = 1
    RULE_residue = 2
    RULE_chemical_shifts = 3
    RULE_chemical_shift = 4
    RULE_number = 5
    RULE_comment = 6

    ruleNames =  [ "olivia_cs", "sequence", "residue", "chemical_shifts", 
                   "chemical_shift", "number", "comment" ]

    EOF = Token.EOF
    Typedef=1
    Separator=2
    Format=3
    Unformat=4
    Eof=5
    Null_string=6
    Integer=7
    Float=8
    Real=9
    COMMENT=10
    SHARP_COMMENT=11
    EXCLM_COMMENT=12
    Simple_name=13
    SPACE=14
    RETURN=15
    SECTION_COMMENT=16
    LINE_COMMENT=17
    Sequence=18
    Ass_tbl_h2o=19
    Ass_tbl_tro=20
    Ass_tbl_d2o=21
    SPACE_TD=22
    RETURN_TD=23
    Tab=24
    Comma=25
    Space=26
    SPACE_SE=27
    RETURN_SE=28
    Chain=29
    Resname=30
    Seqnum=31
    Atomname=32
    Shift=33
    Stddev=34
    SPACE_FO=35
    RETURN_FO=36
    Printf_string=37
    SPACE_PF=38
    RETURN_PF=39
    Any_name=40
    SPACE_CM=41
    RETURN_CM=42

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.0")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class Olivia_csContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(OliviaCSParser.EOF, 0)

        def RETURN(self, i:int=None):
            if i is None:
                return self.getTokens(OliviaCSParser.RETURN)
            else:
                return self.getToken(OliviaCSParser.RETURN, i)

        def comment(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(OliviaCSParser.CommentContext)
            else:
                return self.getTypedRuleContext(OliviaCSParser.CommentContext,i)


        def sequence(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(OliviaCSParser.SequenceContext)
            else:
                return self.getTypedRuleContext(OliviaCSParser.SequenceContext,i)


        def chemical_shifts(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(OliviaCSParser.Chemical_shiftsContext)
            else:
                return self.getTypedRuleContext(OliviaCSParser.Chemical_shiftsContext,i)


        def getRuleIndex(self):
            return OliviaCSParser.RULE_olivia_cs

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOlivia_cs" ):
                listener.enterOlivia_cs(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOlivia_cs" ):
                listener.exitOlivia_cs(self)




    def olivia_cs(self):

        localctx = OliviaCSParser.Olivia_csContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_olivia_cs)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 15
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,0,self._ctx)
            if la_ == 1:
                self.state = 14
                self.match(OliviaCSParser.RETURN)


            self.state = 23
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & 33794) != 0):
                self.state = 21
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,1,self._ctx)
                if la_ == 1:
                    self.state = 17
                    self.comment()
                    pass

                elif la_ == 2:
                    self.state = 18
                    self.sequence()
                    pass

                elif la_ == 3:
                    self.state = 19
                    self.chemical_shifts()
                    pass

                elif la_ == 4:
                    self.state = 20
                    self.match(OliviaCSParser.RETURN)
                    pass


                self.state = 25
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 26
            self.match(OliviaCSParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class SequenceContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Typedef(self):
            return self.getToken(OliviaCSParser.Typedef, 0)

        def Sequence(self):
            return self.getToken(OliviaCSParser.Sequence, 0)

        def RETURN_TD(self):
            return self.getToken(OliviaCSParser.RETURN_TD, 0)

        def Separator(self):
            return self.getToken(OliviaCSParser.Separator, 0)

        def RETURN_SE(self):
            return self.getToken(OliviaCSParser.RETURN_SE, 0)

        def Format(self):
            return self.getToken(OliviaCSParser.Format, 0)

        def Chain(self):
            return self.getToken(OliviaCSParser.Chain, 0)

        def Resname(self):
            return self.getToken(OliviaCSParser.Resname, 0)

        def Seqnum(self):
            return self.getToken(OliviaCSParser.Seqnum, 0)

        def RETURN_FO(self):
            return self.getToken(OliviaCSParser.RETURN_FO, 0)

        def Printf_string(self, i:int=None):
            if i is None:
                return self.getTokens(OliviaCSParser.Printf_string)
            else:
                return self.getToken(OliviaCSParser.Printf_string, i)

        def RETURN_PF(self):
            return self.getToken(OliviaCSParser.RETURN_PF, 0)

        def Unformat(self):
            return self.getToken(OliviaCSParser.Unformat, 0)

        def Tab(self):
            return self.getToken(OliviaCSParser.Tab, 0)

        def Comma(self):
            return self.getToken(OliviaCSParser.Comma, 0)

        def Space(self):
            return self.getToken(OliviaCSParser.Space, 0)

        def residue(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(OliviaCSParser.ResidueContext)
            else:
                return self.getTypedRuleContext(OliviaCSParser.ResidueContext,i)


        def getRuleIndex(self):
            return OliviaCSParser.RULE_sequence

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSequence" ):
                listener.enterSequence(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSequence" ):
                listener.exitSequence(self)




    def sequence(self):

        localctx = OliviaCSParser.SequenceContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_sequence)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 28
            self.match(OliviaCSParser.Typedef)
            self.state = 29
            self.match(OliviaCSParser.Sequence)
            self.state = 30
            self.match(OliviaCSParser.RETURN_TD)
            self.state = 31
            self.match(OliviaCSParser.Separator)
            self.state = 32
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 117440512) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 33
            self.match(OliviaCSParser.RETURN_SE)
            self.state = 34
            self.match(OliviaCSParser.Format)
            self.state = 35
            self.match(OliviaCSParser.Chain)
            self.state = 36
            self.match(OliviaCSParser.Resname)
            self.state = 37
            self.match(OliviaCSParser.Seqnum)
            self.state = 38
            self.match(OliviaCSParser.RETURN_FO)
            self.state = 39
            self.match(OliviaCSParser.Printf_string)
            self.state = 40
            self.match(OliviaCSParser.Printf_string)
            self.state = 41
            self.match(OliviaCSParser.Printf_string)
            self.state = 42
            self.match(OliviaCSParser.RETURN_PF)
            self.state = 44 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 43
                self.residue()
                self.state = 46 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==13):
                    break

            self.state = 48
            self.match(OliviaCSParser.Unformat)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ResidueContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Simple_name(self, i:int=None):
            if i is None:
                return self.getTokens(OliviaCSParser.Simple_name)
            else:
                return self.getToken(OliviaCSParser.Simple_name, i)

        def Integer(self):
            return self.getToken(OliviaCSParser.Integer, 0)

        def RETURN(self):
            return self.getToken(OliviaCSParser.RETURN, 0)

        def getRuleIndex(self):
            return OliviaCSParser.RULE_residue

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterResidue" ):
                listener.enterResidue(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitResidue" ):
                listener.exitResidue(self)




    def residue(self):

        localctx = OliviaCSParser.ResidueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_residue)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 50
            self.match(OliviaCSParser.Simple_name)
            self.state = 51
            self.match(OliviaCSParser.Simple_name)
            self.state = 52
            self.match(OliviaCSParser.Integer)
            self.state = 53
            self.match(OliviaCSParser.RETURN)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Chemical_shiftsContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Typedef(self):
            return self.getToken(OliviaCSParser.Typedef, 0)

        def RETURN_TD(self):
            return self.getToken(OliviaCSParser.RETURN_TD, 0)

        def Separator(self):
            return self.getToken(OliviaCSParser.Separator, 0)

        def RETURN_SE(self):
            return self.getToken(OliviaCSParser.RETURN_SE, 0)

        def Format(self):
            return self.getToken(OliviaCSParser.Format, 0)

        def Chain(self):
            return self.getToken(OliviaCSParser.Chain, 0)

        def Resname(self):
            return self.getToken(OliviaCSParser.Resname, 0)

        def Seqnum(self):
            return self.getToken(OliviaCSParser.Seqnum, 0)

        def Atomname(self):
            return self.getToken(OliviaCSParser.Atomname, 0)

        def Shift(self):
            return self.getToken(OliviaCSParser.Shift, 0)

        def Stddev(self):
            return self.getToken(OliviaCSParser.Stddev, 0)

        def RETURN_FO(self):
            return self.getToken(OliviaCSParser.RETURN_FO, 0)

        def Printf_string(self, i:int=None):
            if i is None:
                return self.getTokens(OliviaCSParser.Printf_string)
            else:
                return self.getToken(OliviaCSParser.Printf_string, i)

        def RETURN_PF(self):
            return self.getToken(OliviaCSParser.RETURN_PF, 0)

        def Unformat(self):
            return self.getToken(OliviaCSParser.Unformat, 0)

        def Ass_tbl_h2o(self):
            return self.getToken(OliviaCSParser.Ass_tbl_h2o, 0)

        def Ass_tbl_tro(self):
            return self.getToken(OliviaCSParser.Ass_tbl_tro, 0)

        def Ass_tbl_d2o(self):
            return self.getToken(OliviaCSParser.Ass_tbl_d2o, 0)

        def Tab(self):
            return self.getToken(OliviaCSParser.Tab, 0)

        def Comma(self):
            return self.getToken(OliviaCSParser.Comma, 0)

        def Space(self):
            return self.getToken(OliviaCSParser.Space, 0)

        def chemical_shift(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(OliviaCSParser.Chemical_shiftContext)
            else:
                return self.getTypedRuleContext(OliviaCSParser.Chemical_shiftContext,i)


        def getRuleIndex(self):
            return OliviaCSParser.RULE_chemical_shifts

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterChemical_shifts" ):
                listener.enterChemical_shifts(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitChemical_shifts" ):
                listener.exitChemical_shifts(self)




    def chemical_shifts(self):

        localctx = OliviaCSParser.Chemical_shiftsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_chemical_shifts)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 55
            self.match(OliviaCSParser.Typedef)
            self.state = 56
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 3670016) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 57
            self.match(OliviaCSParser.RETURN_TD)
            self.state = 58
            self.match(OliviaCSParser.Separator)
            self.state = 59
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 117440512) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 60
            self.match(OliviaCSParser.RETURN_SE)
            self.state = 61
            self.match(OliviaCSParser.Format)
            self.state = 62
            self.match(OliviaCSParser.Chain)
            self.state = 63
            self.match(OliviaCSParser.Resname)
            self.state = 64
            self.match(OliviaCSParser.Seqnum)
            self.state = 65
            self.match(OliviaCSParser.Atomname)
            self.state = 66
            self.match(OliviaCSParser.Shift)
            self.state = 67
            self.match(OliviaCSParser.Stddev)
            self.state = 68
            self.match(OliviaCSParser.RETURN_FO)
            self.state = 69
            self.match(OliviaCSParser.Printf_string)
            self.state = 70
            self.match(OliviaCSParser.Printf_string)
            self.state = 71
            self.match(OliviaCSParser.Printf_string)
            self.state = 72
            self.match(OliviaCSParser.Printf_string)
            self.state = 73
            self.match(OliviaCSParser.Printf_string)
            self.state = 74
            self.match(OliviaCSParser.Printf_string)
            self.state = 75
            self.match(OliviaCSParser.RETURN_PF)
            self.state = 77 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 76
                self.chemical_shift()
                self.state = 79 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==13):
                    break

            self.state = 81
            self.match(OliviaCSParser.Unformat)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Chemical_shiftContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Simple_name(self, i:int=None):
            if i is None:
                return self.getTokens(OliviaCSParser.Simple_name)
            else:
                return self.getToken(OliviaCSParser.Simple_name, i)

        def Integer(self):
            return self.getToken(OliviaCSParser.Integer, 0)

        def number(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(OliviaCSParser.NumberContext)
            else:
                return self.getTypedRuleContext(OliviaCSParser.NumberContext,i)


        def RETURN(self):
            return self.getToken(OliviaCSParser.RETURN, 0)

        def getRuleIndex(self):
            return OliviaCSParser.RULE_chemical_shift

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterChemical_shift" ):
                listener.enterChemical_shift(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitChemical_shift" ):
                listener.exitChemical_shift(self)




    def chemical_shift(self):

        localctx = OliviaCSParser.Chemical_shiftContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_chemical_shift)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 83
            self.match(OliviaCSParser.Simple_name)
            self.state = 84
            self.match(OliviaCSParser.Simple_name)
            self.state = 85
            self.match(OliviaCSParser.Integer)
            self.state = 86
            self.match(OliviaCSParser.Simple_name)
            self.state = 87
            self.number()
            self.state = 88
            self.number()
            self.state = 89
            self.match(OliviaCSParser.RETURN)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class NumberContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Integer(self):
            return self.getToken(OliviaCSParser.Integer, 0)

        def Float(self):
            return self.getToken(OliviaCSParser.Float, 0)

        def Real(self):
            return self.getToken(OliviaCSParser.Real, 0)

        def getRuleIndex(self):
            return OliviaCSParser.RULE_number

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNumber" ):
                listener.enterNumber(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNumber" ):
                listener.exitNumber(self)




    def number(self):

        localctx = OliviaCSParser.NumberContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_number)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 91
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 896) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class CommentContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def COMMENT(self):
            return self.getToken(OliviaCSParser.COMMENT, 0)

        def RETURN_CM(self):
            return self.getToken(OliviaCSParser.RETURN_CM, 0)

        def Any_name(self, i:int=None):
            if i is None:
                return self.getTokens(OliviaCSParser.Any_name)
            else:
                return self.getToken(OliviaCSParser.Any_name, i)

        def getRuleIndex(self):
            return OliviaCSParser.RULE_comment

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterComment" ):
                listener.enterComment(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitComment" ):
                listener.exitComment(self)




    def comment(self):

        localctx = OliviaCSParser.CommentContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_comment)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 93
            self.match(OliviaCSParser.COMMENT)
            self.state = 97
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==40:
                self.state = 94
                self.match(OliviaCSParser.Any_name)
                self.state = 99
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 100
            self.match(OliviaCSParser.RETURN_CM)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








# Generated from BareCSParser.g4 by ANTLR 4.13.0
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,12,48,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,1,0,3,0,10,8,0,1,0,1,0,
        5,0,14,8,0,10,0,12,0,17,9,0,1,0,1,0,1,1,4,1,22,8,1,11,1,12,1,23,
        1,1,4,1,27,8,1,11,1,12,1,28,3,1,31,8,1,1,1,1,1,4,1,35,8,1,11,1,12,
        1,36,1,2,4,2,40,8,2,11,2,12,2,41,1,2,1,2,1,3,1,3,1,3,0,0,4,0,2,4,
        6,0,2,1,1,10,10,2,0,1,2,5,8,51,0,9,1,0,0,0,2,30,1,0,0,0,4,39,1,0,
        0,0,6,45,1,0,0,0,8,10,5,10,0,0,9,8,1,0,0,0,9,10,1,0,0,0,10,15,1,
        0,0,0,11,14,3,2,1,0,12,14,5,10,0,0,13,11,1,0,0,0,13,12,1,0,0,0,14,
        17,1,0,0,0,15,13,1,0,0,0,15,16,1,0,0,0,16,18,1,0,0,0,17,15,1,0,0,
        0,18,19,5,0,0,1,19,1,1,0,0,0,20,22,5,5,0,0,21,20,1,0,0,0,22,23,1,
        0,0,0,23,21,1,0,0,0,23,24,1,0,0,0,24,31,1,0,0,0,25,27,5,6,0,0,26,
        25,1,0,0,0,27,28,1,0,0,0,28,26,1,0,0,0,28,29,1,0,0,0,29,31,1,0,0,
        0,30,21,1,0,0,0,30,26,1,0,0,0,31,32,1,0,0,0,32,34,5,10,0,0,33,35,
        3,4,2,0,34,33,1,0,0,0,35,36,1,0,0,0,36,34,1,0,0,0,36,37,1,0,0,0,
        37,3,1,0,0,0,38,40,3,6,3,0,39,38,1,0,0,0,40,41,1,0,0,0,41,39,1,0,
        0,0,41,42,1,0,0,0,42,43,1,0,0,0,43,44,7,0,0,0,44,5,1,0,0,0,45,46,
        7,1,0,0,46,7,1,0,0,0,8,9,13,15,23,28,30,36,41
    ]

class BareCSParser ( Parser ):

    grammarFileName = "BareCSParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [  ]

    symbolicNames = [ "<INVALID>", "Integer", "Float", "SHARP_COMMENT", 
                      "EXCLM_COMMENT", "Simple_name", "Double_quote_string", 
                      "Double_quote_integer", "Double_quote_float", "SPACE", 
                      "RETURN", "SECTION_COMMENT", "LINE_COMMENT" ]

    RULE_bare_cs = 0
    RULE_cs_raw_format = 1
    RULE_cs_raw_list = 2
    RULE_any = 3

    ruleNames =  [ "bare_cs", "cs_raw_format", "cs_raw_list", "any" ]

    EOF = Token.EOF
    Integer=1
    Float=2
    SHARP_COMMENT=3
    EXCLM_COMMENT=4
    Simple_name=5
    Double_quote_string=6
    Double_quote_integer=7
    Double_quote_float=8
    SPACE=9
    RETURN=10
    SECTION_COMMENT=11
    LINE_COMMENT=12

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.0")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class Bare_csContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(BareCSParser.EOF, 0)

        def RETURN(self, i:int=None):
            if i is None:
                return self.getTokens(BareCSParser.RETURN)
            else:
                return self.getToken(BareCSParser.RETURN, i)

        def cs_raw_format(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(BareCSParser.Cs_raw_formatContext)
            else:
                return self.getTypedRuleContext(BareCSParser.Cs_raw_formatContext,i)


        def getRuleIndex(self):
            return BareCSParser.RULE_bare_cs

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBare_cs" ):
                listener.enterBare_cs(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBare_cs" ):
                listener.exitBare_cs(self)




    def bare_cs(self):

        localctx = BareCSParser.Bare_csContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_bare_cs)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 9
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,0,self._ctx)
            if la_ == 1:
                self.state = 8
                self.match(BareCSParser.RETURN)


            self.state = 15
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & 1120) != 0):
                self.state = 13
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [5, 6]:
                    self.state = 11
                    self.cs_raw_format()
                    pass
                elif token in [10]:
                    self.state = 12
                    self.match(BareCSParser.RETURN)
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 17
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 18
            self.match(BareCSParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Cs_raw_formatContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def RETURN(self):
            return self.getToken(BareCSParser.RETURN, 0)

        def cs_raw_list(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(BareCSParser.Cs_raw_listContext)
            else:
                return self.getTypedRuleContext(BareCSParser.Cs_raw_listContext,i)


        def Simple_name(self, i:int=None):
            if i is None:
                return self.getTokens(BareCSParser.Simple_name)
            else:
                return self.getToken(BareCSParser.Simple_name, i)

        def Double_quote_string(self, i:int=None):
            if i is None:
                return self.getTokens(BareCSParser.Double_quote_string)
            else:
                return self.getToken(BareCSParser.Double_quote_string, i)

        def getRuleIndex(self):
            return BareCSParser.RULE_cs_raw_format

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCs_raw_format" ):
                listener.enterCs_raw_format(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCs_raw_format" ):
                listener.exitCs_raw_format(self)




    def cs_raw_format(self):

        localctx = BareCSParser.Cs_raw_formatContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_cs_raw_format)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 30
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [5]:
                self.state = 21 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while True:
                    self.state = 20
                    self.match(BareCSParser.Simple_name)
                    self.state = 23 
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if not (_la==5):
                        break

                pass
            elif token in [6]:
                self.state = 26 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while True:
                    self.state = 25
                    self.match(BareCSParser.Double_quote_string)
                    self.state = 28 
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if not (_la==6):
                        break

                pass
            else:
                raise NoViableAltException(self)

            self.state = 32
            self.match(BareCSParser.RETURN)
            self.state = 34 
            self._errHandler.sync(self)
            _alt = 1
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 33
                    self.cs_raw_list()

                else:
                    raise NoViableAltException(self)
                self.state = 36 
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,6,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Cs_raw_listContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def RETURN(self):
            return self.getToken(BareCSParser.RETURN, 0)

        def EOF(self):
            return self.getToken(BareCSParser.EOF, 0)

        def any_(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(BareCSParser.AnyContext)
            else:
                return self.getTypedRuleContext(BareCSParser.AnyContext,i)


        def getRuleIndex(self):
            return BareCSParser.RULE_cs_raw_list

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCs_raw_list" ):
                listener.enterCs_raw_list(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCs_raw_list" ):
                listener.exitCs_raw_list(self)




    def cs_raw_list(self):

        localctx = BareCSParser.Cs_raw_listContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_cs_raw_list)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 39 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 38
                self.any_()
                self.state = 41 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not ((((_la) & ~0x3f) == 0 and ((1 << _la) & 486) != 0)):
                    break

            self.state = 43
            _la = self._input.LA(1)
            if not(_la==-1 or _la==10):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AnyContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Float(self):
            return self.getToken(BareCSParser.Float, 0)

        def Integer(self):
            return self.getToken(BareCSParser.Integer, 0)

        def Simple_name(self):
            return self.getToken(BareCSParser.Simple_name, 0)

        def Double_quote_float(self):
            return self.getToken(BareCSParser.Double_quote_float, 0)

        def Double_quote_integer(self):
            return self.getToken(BareCSParser.Double_quote_integer, 0)

        def Double_quote_string(self):
            return self.getToken(BareCSParser.Double_quote_string, 0)

        def getRuleIndex(self):
            return BareCSParser.RULE_any

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAny" ):
                listener.enterAny(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAny" ):
                listener.exitAny(self)




    def any_(self):

        localctx = BareCSParser.AnyContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_any)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 45
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 486) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








# Generated from VnmrPKParser.g4 by ANTLR 4.13.0
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,51,249,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,2,7,7,7,2,8,7,8,2,9,7,9,2,10,7,10,1,0,3,0,24,8,0,1,0,1,0,1,0,1,
        0,4,0,30,8,0,11,0,12,0,31,1,0,4,0,35,8,0,11,0,12,0,36,1,0,4,0,40,
        8,0,11,0,12,0,41,1,0,5,0,45,8,0,10,0,12,0,48,9,0,1,0,1,0,1,1,1,1,
        5,1,54,8,1,10,1,12,1,57,9,1,1,1,1,1,1,2,1,2,1,2,1,2,1,2,3,2,66,8,
        2,1,2,3,2,69,8,2,1,2,1,2,3,2,73,8,2,1,2,3,2,76,8,2,1,2,3,2,79,8,
        2,1,2,3,2,82,8,2,1,2,3,2,85,8,2,1,2,3,2,88,8,2,1,2,3,2,91,8,2,1,
        2,1,2,4,2,95,8,2,11,2,12,2,96,1,2,4,2,100,8,2,11,2,12,2,101,1,2,
        4,2,105,8,2,11,2,12,2,106,3,2,109,8,2,1,3,1,3,1,3,1,3,1,3,3,3,116,
        8,3,1,3,3,3,119,8,3,1,3,3,3,122,8,3,1,3,5,3,125,8,3,10,3,12,3,128,
        9,3,1,3,1,3,1,4,1,4,1,4,1,4,1,4,1,4,3,4,138,8,4,1,4,3,4,141,8,4,
        1,4,3,4,144,8,4,1,4,3,4,147,8,4,1,4,5,4,150,8,4,10,4,12,4,153,9,
        4,1,4,1,4,1,5,1,5,1,5,1,5,1,5,1,5,1,5,3,5,164,8,5,1,5,3,5,167,8,
        5,1,5,3,5,170,8,5,1,5,3,5,173,8,5,1,5,3,5,176,8,5,1,5,5,5,179,8,
        5,10,5,12,5,182,9,5,1,5,1,5,1,6,1,6,1,6,1,6,1,6,1,6,1,6,1,6,1,6,
        3,6,195,8,6,3,6,197,8,6,1,6,1,6,3,6,201,8,6,1,6,3,6,204,8,6,1,6,
        1,6,1,7,1,7,1,7,1,7,1,7,1,7,1,7,3,7,215,8,7,1,7,1,7,1,8,1,8,1,8,
        1,8,1,8,1,8,1,8,1,8,1,8,3,8,228,8,8,1,8,1,8,1,9,1,9,1,9,1,9,1,9,
        1,9,1,9,1,9,1,9,1,9,1,9,3,9,243,8,9,1,9,1,9,1,10,1,10,1,10,0,0,11,
        0,2,4,6,8,10,12,14,16,18,20,0,7,1,1,51,51,2,0,37,37,41,41,2,0,38,
        38,42,42,2,0,39,39,43,43,2,0,40,40,44,44,1,1,14,14,1,0,3,5,285,0,
        23,1,0,0,0,2,51,1,0,0,0,4,60,1,0,0,0,6,110,1,0,0,0,8,131,1,0,0,0,
        10,156,1,0,0,0,12,185,1,0,0,0,14,207,1,0,0,0,16,218,1,0,0,0,18,231,
        1,0,0,0,20,246,1,0,0,0,22,24,5,14,0,0,23,22,1,0,0,0,23,24,1,0,0,
        0,24,46,1,0,0,0,25,45,3,2,1,0,26,45,3,4,2,0,27,45,3,12,6,0,28,30,
        3,14,7,0,29,28,1,0,0,0,30,31,1,0,0,0,31,29,1,0,0,0,31,32,1,0,0,0,
        32,45,1,0,0,0,33,35,3,16,8,0,34,33,1,0,0,0,35,36,1,0,0,0,36,34,1,
        0,0,0,36,37,1,0,0,0,37,45,1,0,0,0,38,40,3,18,9,0,39,38,1,0,0,0,40,
        41,1,0,0,0,41,39,1,0,0,0,41,42,1,0,0,0,42,45,1,0,0,0,43,45,5,14,
        0,0,44,25,1,0,0,0,44,26,1,0,0,0,44,27,1,0,0,0,44,29,1,0,0,0,44,34,
        1,0,0,0,44,39,1,0,0,0,44,43,1,0,0,0,45,48,1,0,0,0,46,44,1,0,0,0,
        46,47,1,0,0,0,47,49,1,0,0,0,48,46,1,0,0,0,49,50,5,0,0,1,50,1,1,0,
        0,0,51,55,5,6,0,0,52,54,5,49,0,0,53,52,1,0,0,0,54,57,1,0,0,0,55,
        53,1,0,0,0,55,56,1,0,0,0,56,58,1,0,0,0,57,55,1,0,0,0,58,59,7,0,0,
        0,59,3,1,0,0,0,60,61,5,2,0,0,61,62,5,30,0,0,62,63,5,31,0,0,63,65,
        5,32,0,0,64,66,5,33,0,0,65,64,1,0,0,0,65,66,1,0,0,0,66,68,1,0,0,
        0,67,69,5,34,0,0,68,67,1,0,0,0,68,69,1,0,0,0,69,70,1,0,0,0,70,72,
        5,35,0,0,71,73,5,36,0,0,72,71,1,0,0,0,72,73,1,0,0,0,73,75,1,0,0,
        0,74,76,7,1,0,0,75,74,1,0,0,0,75,76,1,0,0,0,76,78,1,0,0,0,77,79,
        7,2,0,0,78,77,1,0,0,0,78,79,1,0,0,0,79,81,1,0,0,0,80,82,7,3,0,0,
        81,80,1,0,0,0,81,82,1,0,0,0,82,84,1,0,0,0,83,85,7,4,0,0,84,83,1,
        0,0,0,84,85,1,0,0,0,85,87,1,0,0,0,86,88,5,45,0,0,87,86,1,0,0,0,87,
        88,1,0,0,0,88,90,1,0,0,0,89,91,5,46,0,0,90,89,1,0,0,0,90,91,1,0,
        0,0,91,92,1,0,0,0,92,108,5,48,0,0,93,95,3,6,3,0,94,93,1,0,0,0,95,
        96,1,0,0,0,96,94,1,0,0,0,96,97,1,0,0,0,97,109,1,0,0,0,98,100,3,8,
        4,0,99,98,1,0,0,0,100,101,1,0,0,0,101,99,1,0,0,0,101,102,1,0,0,0,
        102,109,1,0,0,0,103,105,3,10,5,0,104,103,1,0,0,0,105,106,1,0,0,0,
        106,104,1,0,0,0,106,107,1,0,0,0,107,109,1,0,0,0,108,94,1,0,0,0,108,
        99,1,0,0,0,108,104,1,0,0,0,109,5,1,0,0,0,110,111,5,3,0,0,111,112,
        5,4,0,0,112,113,5,4,0,0,113,115,3,20,10,0,114,116,3,20,10,0,115,
        114,1,0,0,0,115,116,1,0,0,0,116,118,1,0,0,0,117,119,3,20,10,0,118,
        117,1,0,0,0,118,119,1,0,0,0,119,121,1,0,0,0,120,122,3,20,10,0,121,
        120,1,0,0,0,121,122,1,0,0,0,122,126,1,0,0,0,123,125,5,7,0,0,124,
        123,1,0,0,0,125,128,1,0,0,0,126,124,1,0,0,0,126,127,1,0,0,0,127,
        129,1,0,0,0,128,126,1,0,0,0,129,130,7,5,0,0,130,7,1,0,0,0,131,132,
        5,3,0,0,132,133,5,4,0,0,133,134,5,4,0,0,134,135,5,4,0,0,135,137,
        3,20,10,0,136,138,3,20,10,0,137,136,1,0,0,0,137,138,1,0,0,0,138,
        140,1,0,0,0,139,141,3,20,10,0,140,139,1,0,0,0,140,141,1,0,0,0,141,
        143,1,0,0,0,142,144,3,20,10,0,143,142,1,0,0,0,143,144,1,0,0,0,144,
        146,1,0,0,0,145,147,3,20,10,0,146,145,1,0,0,0,146,147,1,0,0,0,147,
        151,1,0,0,0,148,150,5,7,0,0,149,148,1,0,0,0,150,153,1,0,0,0,151,
        149,1,0,0,0,151,152,1,0,0,0,152,154,1,0,0,0,153,151,1,0,0,0,154,
        155,7,5,0,0,155,9,1,0,0,0,156,157,5,3,0,0,157,158,5,4,0,0,158,159,
        5,4,0,0,159,160,5,4,0,0,160,161,5,4,0,0,161,163,3,20,10,0,162,164,
        3,20,10,0,163,162,1,0,0,0,163,164,1,0,0,0,164,166,1,0,0,0,165,167,
        3,20,10,0,166,165,1,0,0,0,166,167,1,0,0,0,167,169,1,0,0,0,168,170,
        3,20,10,0,169,168,1,0,0,0,169,170,1,0,0,0,170,172,1,0,0,0,171,173,
        3,20,10,0,172,171,1,0,0,0,172,173,1,0,0,0,173,175,1,0,0,0,174,176,
        3,20,10,0,175,174,1,0,0,0,175,176,1,0,0,0,176,180,1,0,0,0,177,179,
        5,7,0,0,178,177,1,0,0,0,179,182,1,0,0,0,180,178,1,0,0,0,180,181,
        1,0,0,0,181,183,1,0,0,0,182,180,1,0,0,0,183,184,7,5,0,0,184,11,1,
        0,0,0,185,186,5,1,0,0,186,187,5,17,0,0,187,188,5,21,0,0,188,189,
        5,18,0,0,189,196,5,22,0,0,190,191,5,19,0,0,191,194,5,23,0,0,192,
        193,5,20,0,0,193,195,5,24,0,0,194,192,1,0,0,0,194,195,1,0,0,0,195,
        197,1,0,0,0,196,190,1,0,0,0,196,197,1,0,0,0,197,198,1,0,0,0,198,
        200,5,25,0,0,199,201,5,26,0,0,200,199,1,0,0,0,200,201,1,0,0,0,201,
        203,1,0,0,0,202,204,5,27,0,0,203,202,1,0,0,0,203,204,1,0,0,0,204,
        205,1,0,0,0,205,206,5,29,0,0,206,13,1,0,0,0,207,208,5,3,0,0,208,
        209,5,4,0,0,209,210,5,4,0,0,210,211,5,4,0,0,211,212,5,4,0,0,212,
        214,3,20,10,0,213,215,5,10,0,0,214,213,1,0,0,0,214,215,1,0,0,0,215,
        216,1,0,0,0,216,217,7,5,0,0,217,15,1,0,0,0,218,219,5,3,0,0,219,220,
        5,4,0,0,220,221,5,4,0,0,221,222,5,4,0,0,222,223,5,4,0,0,223,224,
        5,4,0,0,224,225,5,4,0,0,225,227,3,20,10,0,226,228,5,11,0,0,227,226,
        1,0,0,0,227,228,1,0,0,0,228,229,1,0,0,0,229,230,7,5,0,0,230,17,1,
        0,0,0,231,232,5,3,0,0,232,233,5,4,0,0,233,234,5,4,0,0,234,235,5,
        4,0,0,235,236,5,4,0,0,236,237,5,4,0,0,237,238,5,4,0,0,238,239,5,
        4,0,0,239,240,5,4,0,0,240,242,3,20,10,0,241,243,5,12,0,0,242,241,
        1,0,0,0,242,243,1,0,0,0,243,244,1,0,0,0,244,245,7,5,0,0,245,19,1,
        0,0,0,246,247,7,6,0,0,247,21,1,0,0,0,42,23,31,36,41,44,46,55,65,
        68,72,75,78,81,84,87,90,96,101,106,108,115,118,121,126,137,140,143,
        146,151,163,166,169,172,175,180,194,196,200,203,214,227,242
    ]

class VnmrPKParser ( Parser ):

    grammarFileName = "VnmrPKParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'peak id.'", "'# Format:'", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "'Dev. 0'", "'Dev. 1'", "'Dev. 2'", "'Dev. 3'", 
                     "<INVALID>", "<INVALID>", "'Assignment'", "<INVALID>", 
                     "<INVALID>", "'Peak_Number'", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "'Label'", "'Comment'" ]

    symbolicNames = [ "<INVALID>", "Peak_id", "Format", "Integer", "Float", 
                      "Real", "COMMENT", "Double_quote_string", "EXCLM_COMMENT", 
                      "SMCLN_COMMENT", "Assignment_2d_ex", "Assignment_3d_ex", 
                      "Assignment_4d_ex", "SPACE", "RETURN", "SECTION_COMMENT", 
                      "LINE_COMMENT", "Dim_0_ppm", "Dim_1_ppm", "Dim_2_ppm", 
                      "Dim_3_ppm", "Dev_0", "Dev_1", "Dev_2", "Dev_3", "Amplitude_LA", 
                      "Volume_LA", "Assignment", "SPACE_LA", "RETURN_LA", 
                      "Peak_number", "X_ppm", "Y_ppm", "Z_ppm", "A_ppm", 
                      "Amplitude", "Volume", "Linewidth_X", "Linewidth_Y", 
                      "Linewidth_Z", "Linewidth_A", "FWHM_X", "FWHM_Y", 
                      "FWHM_Z", "FWHM_A", "Label", "Comment", "SPACE_FO", 
                      "RETURN_FO", "Any_name", "SPACE_CM", "RETURN_CM" ]

    RULE_vnmr_pk = 0
    RULE_comment = 1
    RULE_format = 2
    RULE_peak_ll2d = 3
    RULE_peak_ll3d = 4
    RULE_peak_ll4d = 5
    RULE_data_label = 6
    RULE_peak_2d = 7
    RULE_peak_3d = 8
    RULE_peak_4d = 9
    RULE_number = 10

    ruleNames =  [ "vnmr_pk", "comment", "format", "peak_ll2d", "peak_ll3d", 
                   "peak_ll4d", "data_label", "peak_2d", "peak_3d", "peak_4d", 
                   "number" ]

    EOF = Token.EOF
    Peak_id=1
    Format=2
    Integer=3
    Float=4
    Real=5
    COMMENT=6
    Double_quote_string=7
    EXCLM_COMMENT=8
    SMCLN_COMMENT=9
    Assignment_2d_ex=10
    Assignment_3d_ex=11
    Assignment_4d_ex=12
    SPACE=13
    RETURN=14
    SECTION_COMMENT=15
    LINE_COMMENT=16
    Dim_0_ppm=17
    Dim_1_ppm=18
    Dim_2_ppm=19
    Dim_3_ppm=20
    Dev_0=21
    Dev_1=22
    Dev_2=23
    Dev_3=24
    Amplitude_LA=25
    Volume_LA=26
    Assignment=27
    SPACE_LA=28
    RETURN_LA=29
    Peak_number=30
    X_ppm=31
    Y_ppm=32
    Z_ppm=33
    A_ppm=34
    Amplitude=35
    Volume=36
    Linewidth_X=37
    Linewidth_Y=38
    Linewidth_Z=39
    Linewidth_A=40
    FWHM_X=41
    FWHM_Y=42
    FWHM_Z=43
    FWHM_A=44
    Label=45
    Comment=46
    SPACE_FO=47
    RETURN_FO=48
    Any_name=49
    SPACE_CM=50
    RETURN_CM=51

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.0")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class Vnmr_pkContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(VnmrPKParser.EOF, 0)

        def RETURN(self, i:int=None):
            if i is None:
                return self.getTokens(VnmrPKParser.RETURN)
            else:
                return self.getToken(VnmrPKParser.RETURN, i)

        def comment(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(VnmrPKParser.CommentContext)
            else:
                return self.getTypedRuleContext(VnmrPKParser.CommentContext,i)


        def format_(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(VnmrPKParser.FormatContext)
            else:
                return self.getTypedRuleContext(VnmrPKParser.FormatContext,i)


        def data_label(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(VnmrPKParser.Data_labelContext)
            else:
                return self.getTypedRuleContext(VnmrPKParser.Data_labelContext,i)


        def peak_2d(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(VnmrPKParser.Peak_2dContext)
            else:
                return self.getTypedRuleContext(VnmrPKParser.Peak_2dContext,i)


        def peak_3d(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(VnmrPKParser.Peak_3dContext)
            else:
                return self.getTypedRuleContext(VnmrPKParser.Peak_3dContext,i)


        def peak_4d(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(VnmrPKParser.Peak_4dContext)
            else:
                return self.getTypedRuleContext(VnmrPKParser.Peak_4dContext,i)


        def getRuleIndex(self):
            return VnmrPKParser.RULE_vnmr_pk

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVnmr_pk" ):
                listener.enterVnmr_pk(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVnmr_pk" ):
                listener.exitVnmr_pk(self)




    def vnmr_pk(self):

        localctx = VnmrPKParser.Vnmr_pkContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_vnmr_pk)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 23
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,0,self._ctx)
            if la_ == 1:
                self.state = 22
                self.match(VnmrPKParser.RETURN)


            self.state = 46
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & 16462) != 0):
                self.state = 44
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,4,self._ctx)
                if la_ == 1:
                    self.state = 25
                    self.comment()
                    pass

                elif la_ == 2:
                    self.state = 26
                    self.format_()
                    pass

                elif la_ == 3:
                    self.state = 27
                    self.data_label()
                    pass

                elif la_ == 4:
                    self.state = 29 
                    self._errHandler.sync(self)
                    _alt = 1
                    while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                        if _alt == 1:
                            self.state = 28
                            self.peak_2d()

                        else:
                            raise NoViableAltException(self)
                        self.state = 31 
                        self._errHandler.sync(self)
                        _alt = self._interp.adaptivePredict(self._input,1,self._ctx)

                    pass

                elif la_ == 5:
                    self.state = 34 
                    self._errHandler.sync(self)
                    _alt = 1
                    while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                        if _alt == 1:
                            self.state = 33
                            self.peak_3d()

                        else:
                            raise NoViableAltException(self)
                        self.state = 36 
                        self._errHandler.sync(self)
                        _alt = self._interp.adaptivePredict(self._input,2,self._ctx)

                    pass

                elif la_ == 6:
                    self.state = 39 
                    self._errHandler.sync(self)
                    _alt = 1
                    while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                        if _alt == 1:
                            self.state = 38
                            self.peak_4d()

                        else:
                            raise NoViableAltException(self)
                        self.state = 41 
                        self._errHandler.sync(self)
                        _alt = self._interp.adaptivePredict(self._input,3,self._ctx)

                    pass

                elif la_ == 7:
                    self.state = 43
                    self.match(VnmrPKParser.RETURN)
                    pass


                self.state = 48
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 49
            self.match(VnmrPKParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class CommentContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def COMMENT(self):
            return self.getToken(VnmrPKParser.COMMENT, 0)

        def RETURN_CM(self):
            return self.getToken(VnmrPKParser.RETURN_CM, 0)

        def EOF(self):
            return self.getToken(VnmrPKParser.EOF, 0)

        def Any_name(self, i:int=None):
            if i is None:
                return self.getTokens(VnmrPKParser.Any_name)
            else:
                return self.getToken(VnmrPKParser.Any_name, i)

        def getRuleIndex(self):
            return VnmrPKParser.RULE_comment

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterComment" ):
                listener.enterComment(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitComment" ):
                listener.exitComment(self)




    def comment(self):

        localctx = VnmrPKParser.CommentContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_comment)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 51
            self.match(VnmrPKParser.COMMENT)
            self.state = 55
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==49:
                self.state = 52
                self.match(VnmrPKParser.Any_name)
                self.state = 57
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 58
            _la = self._input.LA(1)
            if not(_la==-1 or _la==51):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class FormatContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Format(self):
            return self.getToken(VnmrPKParser.Format, 0)

        def Peak_number(self):
            return self.getToken(VnmrPKParser.Peak_number, 0)

        def X_ppm(self):
            return self.getToken(VnmrPKParser.X_ppm, 0)

        def Y_ppm(self):
            return self.getToken(VnmrPKParser.Y_ppm, 0)

        def Amplitude(self):
            return self.getToken(VnmrPKParser.Amplitude, 0)

        def RETURN_FO(self):
            return self.getToken(VnmrPKParser.RETURN_FO, 0)

        def Z_ppm(self):
            return self.getToken(VnmrPKParser.Z_ppm, 0)

        def A_ppm(self):
            return self.getToken(VnmrPKParser.A_ppm, 0)

        def Volume(self):
            return self.getToken(VnmrPKParser.Volume, 0)

        def Label(self):
            return self.getToken(VnmrPKParser.Label, 0)

        def Comment(self):
            return self.getToken(VnmrPKParser.Comment, 0)

        def Linewidth_X(self):
            return self.getToken(VnmrPKParser.Linewidth_X, 0)

        def FWHM_X(self):
            return self.getToken(VnmrPKParser.FWHM_X, 0)

        def Linewidth_Y(self):
            return self.getToken(VnmrPKParser.Linewidth_Y, 0)

        def FWHM_Y(self):
            return self.getToken(VnmrPKParser.FWHM_Y, 0)

        def Linewidth_Z(self):
            return self.getToken(VnmrPKParser.Linewidth_Z, 0)

        def FWHM_Z(self):
            return self.getToken(VnmrPKParser.FWHM_Z, 0)

        def Linewidth_A(self):
            return self.getToken(VnmrPKParser.Linewidth_A, 0)

        def FWHM_A(self):
            return self.getToken(VnmrPKParser.FWHM_A, 0)

        def peak_ll2d(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(VnmrPKParser.Peak_ll2dContext)
            else:
                return self.getTypedRuleContext(VnmrPKParser.Peak_ll2dContext,i)


        def peak_ll3d(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(VnmrPKParser.Peak_ll3dContext)
            else:
                return self.getTypedRuleContext(VnmrPKParser.Peak_ll3dContext,i)


        def peak_ll4d(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(VnmrPKParser.Peak_ll4dContext)
            else:
                return self.getTypedRuleContext(VnmrPKParser.Peak_ll4dContext,i)


        def getRuleIndex(self):
            return VnmrPKParser.RULE_format

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterFormat" ):
                listener.enterFormat(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitFormat" ):
                listener.exitFormat(self)




    def format_(self):

        localctx = VnmrPKParser.FormatContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_format)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 60
            self.match(VnmrPKParser.Format)
            self.state = 61
            self.match(VnmrPKParser.Peak_number)
            self.state = 62
            self.match(VnmrPKParser.X_ppm)
            self.state = 63
            self.match(VnmrPKParser.Y_ppm)
            self.state = 65
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==33:
                self.state = 64
                self.match(VnmrPKParser.Z_ppm)


            self.state = 68
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==34:
                self.state = 67
                self.match(VnmrPKParser.A_ppm)


            self.state = 70
            self.match(VnmrPKParser.Amplitude)
            self.state = 72
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==36:
                self.state = 71
                self.match(VnmrPKParser.Volume)


            self.state = 75
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==37 or _la==41:
                self.state = 74
                _la = self._input.LA(1)
                if not(_la==37 or _la==41):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()


            self.state = 78
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==38 or _la==42:
                self.state = 77
                _la = self._input.LA(1)
                if not(_la==38 or _la==42):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()


            self.state = 81
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==39 or _la==43:
                self.state = 80
                _la = self._input.LA(1)
                if not(_la==39 or _la==43):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()


            self.state = 84
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==40 or _la==44:
                self.state = 83
                _la = self._input.LA(1)
                if not(_la==40 or _la==44):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()


            self.state = 87
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==45:
                self.state = 86
                self.match(VnmrPKParser.Label)


            self.state = 90
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==46:
                self.state = 89
                self.match(VnmrPKParser.Comment)


            self.state = 92
            self.match(VnmrPKParser.RETURN_FO)
            self.state = 108
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,19,self._ctx)
            if la_ == 1:
                self.state = 94 
                self._errHandler.sync(self)
                _alt = 1
                while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 93
                        self.peak_ll2d()

                    else:
                        raise NoViableAltException(self)
                    self.state = 96 
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,16,self._ctx)

                pass

            elif la_ == 2:
                self.state = 99 
                self._errHandler.sync(self)
                _alt = 1
                while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 98
                        self.peak_ll3d()

                    else:
                        raise NoViableAltException(self)
                    self.state = 101 
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,17,self._ctx)

                pass

            elif la_ == 3:
                self.state = 104 
                self._errHandler.sync(self)
                _alt = 1
                while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 103
                        self.peak_ll4d()

                    else:
                        raise NoViableAltException(self)
                    self.state = 106 
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,18,self._ctx)

                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Peak_ll2dContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Integer(self):
            return self.getToken(VnmrPKParser.Integer, 0)

        def Float(self, i:int=None):
            if i is None:
                return self.getTokens(VnmrPKParser.Float)
            else:
                return self.getToken(VnmrPKParser.Float, i)

        def number(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(VnmrPKParser.NumberContext)
            else:
                return self.getTypedRuleContext(VnmrPKParser.NumberContext,i)


        def RETURN(self):
            return self.getToken(VnmrPKParser.RETURN, 0)

        def EOF(self):
            return self.getToken(VnmrPKParser.EOF, 0)

        def Double_quote_string(self, i:int=None):
            if i is None:
                return self.getTokens(VnmrPKParser.Double_quote_string)
            else:
                return self.getToken(VnmrPKParser.Double_quote_string, i)

        def getRuleIndex(self):
            return VnmrPKParser.RULE_peak_ll2d

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPeak_ll2d" ):
                listener.enterPeak_ll2d(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPeak_ll2d" ):
                listener.exitPeak_ll2d(self)




    def peak_ll2d(self):

        localctx = VnmrPKParser.Peak_ll2dContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_peak_ll2d)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 110
            self.match(VnmrPKParser.Integer)
            self.state = 111
            self.match(VnmrPKParser.Float)
            self.state = 112
            self.match(VnmrPKParser.Float)
            self.state = 113
            self.number()
            self.state = 115
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,20,self._ctx)
            if la_ == 1:
                self.state = 114
                self.number()


            self.state = 118
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,21,self._ctx)
            if la_ == 1:
                self.state = 117
                self.number()


            self.state = 121
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & 56) != 0):
                self.state = 120
                self.number()


            self.state = 126
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==7:
                self.state = 123
                self.match(VnmrPKParser.Double_quote_string)
                self.state = 128
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 129
            _la = self._input.LA(1)
            if not(_la==-1 or _la==14):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Peak_ll3dContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Integer(self):
            return self.getToken(VnmrPKParser.Integer, 0)

        def Float(self, i:int=None):
            if i is None:
                return self.getTokens(VnmrPKParser.Float)
            else:
                return self.getToken(VnmrPKParser.Float, i)

        def number(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(VnmrPKParser.NumberContext)
            else:
                return self.getTypedRuleContext(VnmrPKParser.NumberContext,i)


        def RETURN(self):
            return self.getToken(VnmrPKParser.RETURN, 0)

        def EOF(self):
            return self.getToken(VnmrPKParser.EOF, 0)

        def Double_quote_string(self, i:int=None):
            if i is None:
                return self.getTokens(VnmrPKParser.Double_quote_string)
            else:
                return self.getToken(VnmrPKParser.Double_quote_string, i)

        def getRuleIndex(self):
            return VnmrPKParser.RULE_peak_ll3d

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPeak_ll3d" ):
                listener.enterPeak_ll3d(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPeak_ll3d" ):
                listener.exitPeak_ll3d(self)




    def peak_ll3d(self):

        localctx = VnmrPKParser.Peak_ll3dContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_peak_ll3d)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 131
            self.match(VnmrPKParser.Integer)
            self.state = 132
            self.match(VnmrPKParser.Float)
            self.state = 133
            self.match(VnmrPKParser.Float)
            self.state = 134
            self.match(VnmrPKParser.Float)
            self.state = 135
            self.number()
            self.state = 137
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,24,self._ctx)
            if la_ == 1:
                self.state = 136
                self.number()


            self.state = 140
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,25,self._ctx)
            if la_ == 1:
                self.state = 139
                self.number()


            self.state = 143
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,26,self._ctx)
            if la_ == 1:
                self.state = 142
                self.number()


            self.state = 146
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & 56) != 0):
                self.state = 145
                self.number()


            self.state = 151
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==7:
                self.state = 148
                self.match(VnmrPKParser.Double_quote_string)
                self.state = 153
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 154
            _la = self._input.LA(1)
            if not(_la==-1 or _la==14):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Peak_ll4dContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Integer(self):
            return self.getToken(VnmrPKParser.Integer, 0)

        def Float(self, i:int=None):
            if i is None:
                return self.getTokens(VnmrPKParser.Float)
            else:
                return self.getToken(VnmrPKParser.Float, i)

        def number(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(VnmrPKParser.NumberContext)
            else:
                return self.getTypedRuleContext(VnmrPKParser.NumberContext,i)


        def RETURN(self):
            return self.getToken(VnmrPKParser.RETURN, 0)

        def EOF(self):
            return self.getToken(VnmrPKParser.EOF, 0)

        def Double_quote_string(self, i:int=None):
            if i is None:
                return self.getTokens(VnmrPKParser.Double_quote_string)
            else:
                return self.getToken(VnmrPKParser.Double_quote_string, i)

        def getRuleIndex(self):
            return VnmrPKParser.RULE_peak_ll4d

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPeak_ll4d" ):
                listener.enterPeak_ll4d(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPeak_ll4d" ):
                listener.exitPeak_ll4d(self)




    def peak_ll4d(self):

        localctx = VnmrPKParser.Peak_ll4dContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_peak_ll4d)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 156
            self.match(VnmrPKParser.Integer)
            self.state = 157
            self.match(VnmrPKParser.Float)
            self.state = 158
            self.match(VnmrPKParser.Float)
            self.state = 159
            self.match(VnmrPKParser.Float)
            self.state = 160
            self.match(VnmrPKParser.Float)
            self.state = 161
            self.number()
            self.state = 163
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,29,self._ctx)
            if la_ == 1:
                self.state = 162
                self.number()


            self.state = 166
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,30,self._ctx)
            if la_ == 1:
                self.state = 165
                self.number()


            self.state = 169
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,31,self._ctx)
            if la_ == 1:
                self.state = 168
                self.number()


            self.state = 172
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,32,self._ctx)
            if la_ == 1:
                self.state = 171
                self.number()


            self.state = 175
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & 56) != 0):
                self.state = 174
                self.number()


            self.state = 180
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==7:
                self.state = 177
                self.match(VnmrPKParser.Double_quote_string)
                self.state = 182
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 183
            _la = self._input.LA(1)
            if not(_la==-1 or _la==14):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Data_labelContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Peak_id(self):
            return self.getToken(VnmrPKParser.Peak_id, 0)

        def Dim_0_ppm(self):
            return self.getToken(VnmrPKParser.Dim_0_ppm, 0)

        def Dev_0(self):
            return self.getToken(VnmrPKParser.Dev_0, 0)

        def Dim_1_ppm(self):
            return self.getToken(VnmrPKParser.Dim_1_ppm, 0)

        def Dev_1(self):
            return self.getToken(VnmrPKParser.Dev_1, 0)

        def Amplitude_LA(self):
            return self.getToken(VnmrPKParser.Amplitude_LA, 0)

        def RETURN_LA(self):
            return self.getToken(VnmrPKParser.RETURN_LA, 0)

        def Dim_2_ppm(self):
            return self.getToken(VnmrPKParser.Dim_2_ppm, 0)

        def Dev_2(self):
            return self.getToken(VnmrPKParser.Dev_2, 0)

        def Volume_LA(self):
            return self.getToken(VnmrPKParser.Volume_LA, 0)

        def Assignment(self):
            return self.getToken(VnmrPKParser.Assignment, 0)

        def Dim_3_ppm(self):
            return self.getToken(VnmrPKParser.Dim_3_ppm, 0)

        def Dev_3(self):
            return self.getToken(VnmrPKParser.Dev_3, 0)

        def getRuleIndex(self):
            return VnmrPKParser.RULE_data_label

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterData_label" ):
                listener.enterData_label(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitData_label" ):
                listener.exitData_label(self)




    def data_label(self):

        localctx = VnmrPKParser.Data_labelContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_data_label)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 185
            self.match(VnmrPKParser.Peak_id)
            self.state = 186
            self.match(VnmrPKParser.Dim_0_ppm)
            self.state = 187
            self.match(VnmrPKParser.Dev_0)
            self.state = 188
            self.match(VnmrPKParser.Dim_1_ppm)
            self.state = 189
            self.match(VnmrPKParser.Dev_1)
            self.state = 196
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==19:
                self.state = 190
                self.match(VnmrPKParser.Dim_2_ppm)
                self.state = 191
                self.match(VnmrPKParser.Dev_2)
                self.state = 194
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==20:
                    self.state = 192
                    self.match(VnmrPKParser.Dim_3_ppm)
                    self.state = 193
                    self.match(VnmrPKParser.Dev_3)




            self.state = 198
            self.match(VnmrPKParser.Amplitude_LA)
            self.state = 200
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==26:
                self.state = 199
                self.match(VnmrPKParser.Volume_LA)


            self.state = 203
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==27:
                self.state = 202
                self.match(VnmrPKParser.Assignment)


            self.state = 205
            self.match(VnmrPKParser.RETURN_LA)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Peak_2dContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Integer(self):
            return self.getToken(VnmrPKParser.Integer, 0)

        def Float(self, i:int=None):
            if i is None:
                return self.getTokens(VnmrPKParser.Float)
            else:
                return self.getToken(VnmrPKParser.Float, i)

        def number(self):
            return self.getTypedRuleContext(VnmrPKParser.NumberContext,0)


        def RETURN(self):
            return self.getToken(VnmrPKParser.RETURN, 0)

        def EOF(self):
            return self.getToken(VnmrPKParser.EOF, 0)

        def Assignment_2d_ex(self):
            return self.getToken(VnmrPKParser.Assignment_2d_ex, 0)

        def getRuleIndex(self):
            return VnmrPKParser.RULE_peak_2d

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPeak_2d" ):
                listener.enterPeak_2d(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPeak_2d" ):
                listener.exitPeak_2d(self)




    def peak_2d(self):

        localctx = VnmrPKParser.Peak_2dContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_peak_2d)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 207
            self.match(VnmrPKParser.Integer)
            self.state = 208
            self.match(VnmrPKParser.Float)
            self.state = 209
            self.match(VnmrPKParser.Float)
            self.state = 210
            self.match(VnmrPKParser.Float)
            self.state = 211
            self.match(VnmrPKParser.Float)
            self.state = 212
            self.number()
            self.state = 214
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==10:
                self.state = 213
                self.match(VnmrPKParser.Assignment_2d_ex)


            self.state = 216
            _la = self._input.LA(1)
            if not(_la==-1 or _la==14):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Peak_3dContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Integer(self):
            return self.getToken(VnmrPKParser.Integer, 0)

        def Float(self, i:int=None):
            if i is None:
                return self.getTokens(VnmrPKParser.Float)
            else:
                return self.getToken(VnmrPKParser.Float, i)

        def number(self):
            return self.getTypedRuleContext(VnmrPKParser.NumberContext,0)


        def RETURN(self):
            return self.getToken(VnmrPKParser.RETURN, 0)

        def EOF(self):
            return self.getToken(VnmrPKParser.EOF, 0)

        def Assignment_3d_ex(self):
            return self.getToken(VnmrPKParser.Assignment_3d_ex, 0)

        def getRuleIndex(self):
            return VnmrPKParser.RULE_peak_3d

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPeak_3d" ):
                listener.enterPeak_3d(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPeak_3d" ):
                listener.exitPeak_3d(self)




    def peak_3d(self):

        localctx = VnmrPKParser.Peak_3dContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_peak_3d)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 218
            self.match(VnmrPKParser.Integer)
            self.state = 219
            self.match(VnmrPKParser.Float)
            self.state = 220
            self.match(VnmrPKParser.Float)
            self.state = 221
            self.match(VnmrPKParser.Float)
            self.state = 222
            self.match(VnmrPKParser.Float)
            self.state = 223
            self.match(VnmrPKParser.Float)
            self.state = 224
            self.match(VnmrPKParser.Float)
            self.state = 225
            self.number()
            self.state = 227
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==11:
                self.state = 226
                self.match(VnmrPKParser.Assignment_3d_ex)


            self.state = 229
            _la = self._input.LA(1)
            if not(_la==-1 or _la==14):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Peak_4dContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Integer(self):
            return self.getToken(VnmrPKParser.Integer, 0)

        def Float(self, i:int=None):
            if i is None:
                return self.getTokens(VnmrPKParser.Float)
            else:
                return self.getToken(VnmrPKParser.Float, i)

        def number(self):
            return self.getTypedRuleContext(VnmrPKParser.NumberContext,0)


        def RETURN(self):
            return self.getToken(VnmrPKParser.RETURN, 0)

        def EOF(self):
            return self.getToken(VnmrPKParser.EOF, 0)

        def Assignment_4d_ex(self):
            return self.getToken(VnmrPKParser.Assignment_4d_ex, 0)

        def getRuleIndex(self):
            return VnmrPKParser.RULE_peak_4d

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPeak_4d" ):
                listener.enterPeak_4d(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPeak_4d" ):
                listener.exitPeak_4d(self)




    def peak_4d(self):

        localctx = VnmrPKParser.Peak_4dContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_peak_4d)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 231
            self.match(VnmrPKParser.Integer)
            self.state = 232
            self.match(VnmrPKParser.Float)
            self.state = 233
            self.match(VnmrPKParser.Float)
            self.state = 234
            self.match(VnmrPKParser.Float)
            self.state = 235
            self.match(VnmrPKParser.Float)
            self.state = 236
            self.match(VnmrPKParser.Float)
            self.state = 237
            self.match(VnmrPKParser.Float)
            self.state = 238
            self.match(VnmrPKParser.Float)
            self.state = 239
            self.match(VnmrPKParser.Float)
            self.state = 240
            self.number()
            self.state = 242
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==12:
                self.state = 241
                self.match(VnmrPKParser.Assignment_4d_ex)


            self.state = 244
            _la = self._input.LA(1)
            if not(_la==-1 or _la==14):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class NumberContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Real(self):
            return self.getToken(VnmrPKParser.Real, 0)

        def Float(self):
            return self.getToken(VnmrPKParser.Float, 0)

        def Integer(self):
            return self.getToken(VnmrPKParser.Integer, 0)

        def getRuleIndex(self):
            return VnmrPKParser.RULE_number

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNumber" ):
                listener.enterNumber(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNumber" ):
                listener.exitNumber(self)




    def number(self):

        localctx = VnmrPKParser.NumberContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_number)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 246
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 56) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








##
# File: CifToNmrStar.py
# Date: 19-Jul-2021
#
# Updates:
# 13-Oct-2021  M. Yokochi - code revision according to PEP8 using Pylint (DAOTHER-7389, issue #5)
# 20-Apr-2022  M. Yokochi - enable to fix broken datablock order of CIF formatted NMR-STAR using NMR-STAR schema (DAOTHER-7407, NMR restraint remediation)
# 28-Jul-2022  M. Yokochi - enable to fix format issue of CIF formatted NMR-STAR (You cannot have two loops with the same category in one saveframe. Category: '_Audit')
# 27-Sep-2022  M. Yokochi - auto fill list ID and entry ID (NMR restraint remediation)
# 13-Jun-2023  M. Yokochi - sort loops in a saveframe based on schema
# 30-May-2024  M. Yokochi - resolve duplication of datablock/saveframe name (DAOTHER-9437)
# 25-Jun-2024  M. Yokochi - strip white spaces in a datablock name derived from the model file (DAOTHER-9511)
# 07-Jan-2025  M. Yokochi - retrieve symbolic label representations (DAOTHER-1728, 9846)
##
""" Wrapper class for CIF to NMR-STAR converter.
    @author: Masashi Yokochi
"""
__docformat__ = "restructuredtext en"
__author__ = "Masashi Yokochi"
__email__ = "yokochi@protein.osaka-u.ac.jp"
__license__ = "Apache License 2.0"
__version__ = "1.0.0"

import sys
import os
import re
import pynmrstar
import pickle
import logging
import hashlib
import collections

from packaging import version
from operator import itemgetter
from typing import Any, IO, Union, Optional

try:
    from wwpdb.utils.nmr.io.mmCIFUtil import mmCIFUtil
    from wwpdb.utils.nmr.AlignUtil import (emptyValue,
                                           trueValue)
except ImportError:
    from nmr.io.mmCIFUtil import mmCIFUtil
    from nmr.AlignUtil import (emptyValue,
                               trueValue)


__pynmrstar_v3_3_1__ = version.parse(pynmrstar.__version__) >= version.parse("3.3.1")


if __pynmrstar_v3_3_1__:
    logger = logging.getLogger('pynmrstar')
    logger.setLevel(logging.ERROR)
else:
    logging.getLogger().setLevel(logging.ERROR)  # set level for pynmrstar


def has_key_value(d: dict, key: Any) -> bool:
    """ Return whether a given dictionary has effective value for a key.
        @return: True if d[key] has effective value, False otherwise
    """

    if not isinstance(d, dict) or key is None:
        return False

    if key in d:
        return d[key] is not None

    return False


def get_value_safe(d: Optional[Union[dict, list, tuple]] = None, key: Optional = None, default: Optional = None) -> Any:
    """ Return value of a given dictionary, list, or tuple for a key.
        @return: value for a key, None (by default) otherwise
    """

    if None in (d, key):
        return default

    return d.get(key, default)


def get_first_sf_tag(sf: pynmrstar.Saveframe, tag: str, default: str = '') -> Any:
    """ Return the first value of a given saveframe tag with decoding symbol notation.
        @return: The first tag value, '' (by default) otherwise.
    """

    if not isinstance(sf, pynmrstar.Saveframe) or tag is None:
        return default

    array = sf.get_tag(tag)

    if len(array) == 0 or array[0] is None:
        return default

    if not isinstance(array[0], str):
        return array[0]

    value = array[0]

    while value.startswith('$$'):
        value = value[1:]

    if len(value) == 0 or value == '$':
        return default

    return value if len(value) < 2 or value[0] != '$' else value[1:]


def set_sf_tag(sf: pynmrstar.Saveframe, tag: str, value: Any):
    """ Set saveframe tag with a given value.
    """

    tagNames = [t[0] for t in sf.tags]

    if isinstance(value, str):

        if len(value) == 0:
            value = None

        if value is not None:
            while value.startswith('$$'):
                value = value[1:]

            if len(value) == 0 or value == '$':
                value = None

    if tag not in tagNames:
        sf.add_tag(tag, value)
        return

    sf.tags[tagNames.index(tag)][1] = value


def set_lp_tag(lp: pynmrstar.Loop, tag: str, value: Any):
    """ Set loop tag with a given value.
    """

    if tag not in lp.tags:
        lp.add_tag(tag)

        for row in lp:
            row.append(value)

    else:
        col = lp.tags.index(tag)

        for row in lp:
            row[col] = value


def retrieve_symbolic_labels(strData: pynmrstar.Entry):
    """ Retrieve symbolic label representations that serve as saveframe pointers in NMR-STAR.
    """

    def get_parent_sf_framecode(parent_sf_tag_prefix, parent_list_id):
        if isinstance(parent_list_id, int) or (isinstance(parent_list_id, str) and parent_list_id.isdigit()):
            try:
                parent_sf = strData.get_saveframes_by_tag_and_value(f'{parent_sf_tag_prefix}.ID', parent_list_id)[0]
                return get_first_sf_tag(parent_sf, 'Sf_framecode')
            except IndexError:
                try:
                    parent_sf = strData.get_saveframes_by_tag_and_value(f'{parent_sf_tag_prefix}.ID', int(parent_list_id)
                                                                        if isinstance(parent_list_id, str) else str(parent_list_id))[0]
                    return get_first_sf_tag(parent_sf, 'Sf_framecode')
                except IndexError:
                    pass
        return ''

    for sf in strData.frame_list:
        for idx, tag in enumerate(sf.tags):
            if tag[0].endswith('_label'):
                if tag[1] not in emptyValue:
                    if not tag[1].startswith('$'):
                        sf.tags[idx][1] = '$' + tag[1]
                else:
                    id_tag = tag[0][:-6] + '_ID'
                    id_val = get_first_sf_tag(sf, id_tag)
                    if id_val not in emptyValue:
                        parent_sf_framecode = get_parent_sf_framecode(f'_{tag[0][:-6]}', id_val)
                        if len(parent_sf_framecode) > 0 and id_tag in sf.tags:
                            set_sf_tag(sf, id_tag, f'${parent_sf_framecode}')

        for lp in sf.loops:
            label_cols = [idx for idx, tag in enumerate(lp.tags) if tag.endswith('_label')]
            if len(label_cols) == 0:
                continue
            id_cols = [-1] * len(label_cols)
            for idx, label_col in enumerate(label_cols):
                id_tag = lp.tags[label_col][:-6] + '_ID'
                if id_tag in lp.tags:
                    id_cols[idx] = lp.tags.index(id_tag)
            for idx, row in enumerate(lp.data):
                for col, val in enumerate(row):
                    if col in label_cols:
                        if val not in emptyValue:
                            if not val.startswith('$'):
                                lp.data[idx][col] = '$' + val
                        else:
                            id_col = next((id_col for label_col, id_col in zip(label_cols, id_cols) if col == label_col), -1)
                            if id_col == -1 or row[id_col] in emptyValue:
                                continue
                            parent_sf_framecode = get_parent_sf_framecode(f'_{lp.tags[col][:-6]}', row[id_col])
                            if len(parent_sf_framecode) > 0:
                                lp.data[idx][col] = f'${parent_sf_framecode}'


class CifToNmrStar:
    """ Simple CIF to NMR-STAR converter.
    """

    def __init__(self, log: IO = sys.stderr):
        self.__class_name__ = self.__class__.__name__
        self.__version__ = __version__

        self.__lfh = log

        # directory
        self.schema_dir = os.path.dirname(__file__) + '/nmr-star_schema/'

        def load_schema_from_pickle(file_name):
            """ Load NMR-STAR schema from pickle file.
            """

            if os.path.exists(file_name):

                with open(file_name, 'rb') as ifh:
                    return pickle.load(ifh)

            return None

        # NMR-STAR schema
        self.schema = load_schema_from_pickle(self.schema_dir + 'schema.pkl')

        # NMR-STAR schema order
        # self.schema_order = load_schema_from_pickle(self.schema_dir + 'schema_order.pkl')

        # NMR-STAR category order
        self.category_order = load_schema_from_pickle(self.schema_dir + 'category_order.pkl')

        if self.schema is None:
            schema = pynmrstar.Schema()  # network latency occurs if pickle resource files are not available
            self.schema = schema.schema
            self.category_order = schema.category_order

        # NEF category order
        self.category_order_nef = ['_nef_nmr_meta_data',
                                   '_nef_molecular_system',
                                   '_nef_chemical_shift_list',
                                   '_nef_distance_restraint_list',
                                   '_nef_dihedral_restraint_list',
                                   '_nef_rdc_restraint_list',
                                   '_nef_nmr_spectrum',
                                   '_nef_peak_restraint_links']

    def write_schema_as_pickles(self):
        """ Retrieve NMR-STAR schema from pynmrstar.Schema, then write schema objects as each pickle file.
        """

        def write_schema_as_pickle(obj, file_name):
            """ Write NMR-STAR schema as pickle file.
            """

            with open(file_name, 'wb') as ofh:
                pickle.dump(obj, ofh)

        schema = pynmrstar.Schema()  # retrieve the latest schema via interet access

        # print(schema.headers)
        with open(self.schema_dir + 'headers.txt', 'w') as ofh:
            for header in schema.headers:
                ofh.write(header + '\n')
        self.__lfh.write('headers.txt: Done.\n')

        # print(schema.schema)
        write_schema_as_pickle(schema.schema, self.schema_dir + 'schema.pkl')
        self.__lfh.write('schema.pkl: Done.\n')

        # print(schema.schema_order)
        write_schema_as_pickle(schema.schema_order, self.schema_dir + 'schema_order.pkl')
        self.__lfh.write('schema_order.pkl: Done.\n')

        # print(schema.category_order)
        write_schema_as_pickle(schema.category_order, self.schema_dir + 'category_order.pkl')
        self.__lfh.write('category_order.pkl: Done.\n')

        # print(schema.data_types)
        write_schema_as_pickle(schema.data_types, self.schema_dir + 'data_types.pkl')
        self.__lfh.write('data_types.pkl: Done.\n')

        with open(self.schema_dir + 'version.txt', 'w') as ofh:
            ofh.write(schema.version)
        self.__lfh.write(f"version: {schema.version}\n")

    def convert(self, cifPath: Optional[str] = None, strPath: Optional[str] = None,
                datablockName: Optional[str] = None, originalFileName: Optional[str] = None, maxRepeat: int = 1) -> bool:
        """ Convert CIF formatted NMR data file to normalized NMR-STAR file.
        """

        if None in (cifPath, strPath):
            return False

        try:

            cifObj = mmCIFUtil(filePath=cifPath)

            block_name_list = cifObj.getDataBlockNameList()

            if len(block_name_list) == 0:  # single loop

                if maxRepeat != 1:
                    return False

                datablock_pattern = re.compile(r'\s*data_(\S+)\s*')
                sf_anonymous_pattern = re.compile(r'\s*save_\S+\s*')

                has_datablock = has_anonymous_saveframe = False

                with open(cifPath, 'r', encoding='utf-8') as ifh:
                    for line in ifh:
                        if datablock_pattern.match(line):
                            has_datablock = True
                        elif sf_anonymous_pattern.match(line):
                            has_anonymous_saveframe = True
                            break

                if has_datablock or not has_anonymous_saveframe:
                    return False

                with open(cifPath, 'r', encoding='utf-8') as ifh, \
                        open(cifPath + '~', 'w', encoding='utf-8') as ofh:
                    name = datablockName
                    if datablockName is None:
                        name = originalFileName
                    if datablockName is None:
                        name = os.path.basename(cifPath)
                    ofh.write('data_' + name + '\n\n')
                    for line in ifh:
                        ofh.write(line)

                    os.replace(cifPath + '~', cifPath)

                return self.convert(cifPath, strPath, datablockName, originalFileName, maxRepeat - 1)

            dup_block_name_list = []
            if len(block_name_list) > 1:
                c = collections.Counter(block_name_list).most_common()
                dup_block_name_list = [k for k, v in c if v > 1]

            block_name_pat_w_list_0 = [block_name[:-1] for block_name in block_name_list if block_name.endswith('list_0')]

            block_name_counter = {n: 0 for n in block_name_list}

            entry_id = None

            strData = pynmrstar.Entry.from_scratch(datablockName if datablockName is not None else os.path.basename(cifPath))

            # check category order in CIF
            category_order = []
            previous_order = -1

            sf_category_counter = {}

            for block_name in block_name_list:
                block_name_counter[block_name] += 1
                ext = block_name_counter[block_name]

                dBlockStruct = cifObj.getDataBlockStructure(block_name, ext)

                ordered_block = True
                sf_category = ''

                for category, itVals in dBlockStruct.items():
                    try:
                        current_order = self.category_order.index('_' + category)
                    except ValueError:
                        continue
                    # print(f"{block_name} {category} {current_order}")
                    item = {'block_name': block_name, 'category': category,
                            'category_order': current_order,
                            'block_name_ext': ext}
                    if current_order < previous_order:
                        ordered_block = False
                        for item2 in category_order:
                            if item2['block_name'] == block_name:
                                item2['ordered'] = False
                    item['ordered'] = ordered_block

                    item['sf_category_flag'] = False
                    for _item, _value in zip(itVals['Items'], itVals['Values'][0]):
                        tag = '_' + category + '.' + _item
                        tag = tag.lower()
                        if tag in self.schema:
                            sdict = {k: v for k, v in self.schema[tag].items() if v not in emptyValue}
                            if 'sf_category' not in item:
                                item['sf_category'] = sdict['SFCategory']
                            if 'super_category' not in item:
                                item['super_category'] = sdict['ADIT super category']
                            if 'Sf category flag' in sdict and sdict['Sf category flag'].lower() in trueValue:
                                item['sf_category_flag'] = True
                            if _item == 'Entry_ID' and entry_id in emptyValue:
                                entry_id = _value

                    if 'Sf_category' not in itVals['Items']:
                        _tag = '_' + category.lower() + '.sf_category'
                        if _tag in self.schema:
                            sdict = {k: v for k, v in self.schema[_tag].items() if v not in emptyValue}
                            if 'sf_category' not in item:
                                item['sf_category'] = sdict['SFCategory']
                            if 'super_category' not in item:
                                item['super_category'] = sdict['ADIT super category']
                            if 'Sf category flag' in sdict and sdict['Sf category flag'].lower() in trueValue:
                                item['sf_category_flag'] = True

                    if len(sf_category) == 0:
                        sf_category = item['sf_category']
                    elif sf_category != item['sf_category']:
                        for item2 in category_order:
                            if item2['block_name'] == block_name:
                                item2['ordered'] = False
                        item['ordered'] = False

                    if item not in category_order:
                        category_order.append(item)

                    if item['sf_category_flag']:
                        previous_order = current_order

            if entry_id in emptyValue:
                entry_id = block_name_list[0].strip().replace(' ', '_')  # DAOTHER-9511: replace white space in a datablock name to underscore

            _entry_id = entry_id.upper()

            if datablockName is None:
                strData.entry_id = f'nmr_data_{entry_id.lower()}'

            # reorder
            category_order.sort(key=itemgetter('category_order'))

            # correct block name in case
            reserved_block_names = []
            prev_sf_category = ''
            prev_sf_block_name = ''
            for item in category_order:
                block_name = item['block_name']
                split_block_name = block_name.split('_')

                if block_name in dup_block_name_list:
                    if split_block_name[-1].isdigit():
                        split_block_name.pop()
                    item['new_block_name'] = '_'.join(split_block_name) + '_' + str(item['block_name_ext'])

                elif any(block_name.startswith(pat) for pat in block_name_pat_w_list_0):  # STARCh output
                    if split_block_name[-1].isdigit():
                        list_id = int(split_block_name[-1]) + 1
                        split_block_name.pop()
                        item['new_block_name'] = '_'.join(split_block_name) + '_' + str(list_id)

                if item['ordered']:
                    continue

                if item['sf_category_flag']:
                    if item['sf_category'] == item['super_category']:
                        block_name = item['sf_category']
                    else:
                        list_id = 1
                        while True:
                            block_name = item['sf_category'] + '_' + str(list_id)
                            if block_name not in reserved_block_names:
                                break
                            list_id += 1
                    prev_sf_category = item['sf_category']
                    prev_sf_block_name = block_name
                else:
                    if item['sf_category'] == prev_sf_category:
                        block_name = prev_sf_block_name
                    else:
                        item['missing_sf_category'] = True
                        if item['sf_category'] == item['super_category']:
                            block_name = item['sf_category']
                        else:
                            list_id = 1
                            while True:
                                block_name = item['sf_category'] + '_' + str(list_id)
                                if block_name not in reserved_block_names:
                                    break
                                list_id += 1
                        prev_sf_category = item['sf_category']
                        prev_sf_block_name = block_name

                if 'new_block_name' not in item:
                    item['new_block_name'] = block_name

            ordered_block_names = []
            for item in category_order:
                block_name = item['new_block_name'] if 'new_block_name' in item else item['block_name']
                if block_name not in ordered_block_names:
                    ordered_block_names.append(block_name)

            _category_order = []
            for _block_name in ordered_block_names:
                count = 0
                for item in category_order:
                    block_name = item['new_block_name'] if 'new_block_name' in item else item['block_name']
                    if block_name == _block_name:
                        _category_order.append(item)
                        count += 1
                if count == 1:
                    last_item = _category_order[-1]
                    if not last_item['sf_category_flag'] and last_item['ordered']:
                        last_item['missing_sf_category'] = True
                        last_item['new_block_name'] = last_item['block_name']

            category_order = _category_order

            _sf_category = ''
            cur_list_id = 1

            reserved_block_names = []
            sf = None
            for item in category_order:
                block_name = item['block_name']
                category = item['category']
                sf_category = item['sf_category']
                ext = item['block_name_ext']

                if item['sf_category_flag'] or ('missing_sf_category' in item and item['missing_sf_category']):
                    if sf_category not in sf_category_counter:
                        cur_list_id = 1
                    else:
                        cur_list_id = sf_category_counter[sf_category] + 1
                elif sf_category != _sf_category:
                    cur_list_id = 1

                _sf_category = sf_category

                sf_tag_prefix = next(v['Tag category'] for k, v in self.schema.items()
                                     if v['SFCategory'] == sf_category and v['Tag field'] == 'Sf_category')

                sf_category_counter[sf_category] = cur_list_id
                _cur_list_id = str(cur_list_id)

                if 'missing_sf_category' in item and item['missing_sf_category']:
                    new_block_name = item['new_block_name']

                    if new_block_name in reserved_block_names:
                        continue

                    if sf is not None:
                        strData.add_saveframe(sf)

                    sf = pynmrstar.Saveframe.from_scratch(new_block_name)
                    reserved_block_names.append(new_block_name)
                    sf.set_tag_prefix(sf_tag_prefix)
                    sf.add_tag('Sf_category', sf_category)
                    sf.add_tag('Sf_framecode', new_block_name)
                    sf.add_tag('Entry_ID', _entry_id)

                    if sf_category != item['super_category']:
                        sf.add_tag('ID', _cur_list_id)

                elif item['sf_category_flag']:
                    new_block_name = item['new_block_name'] if 'new_block_name' in item else block_name

                    if new_block_name in reserved_block_names:
                        continue

                    if sf is not None:
                        strData.add_saveframe(sf)

                    sf = pynmrstar.Saveframe.from_scratch(new_block_name)
                    reserved_block_names.append(new_block_name)
                    sf.set_tag_prefix(category)

                    dBlockStruct = cifObj.getDataBlockStructure(block_name, ext)
                    itVals = next(v for k, v in dBlockStruct.items() if k == category)

                    sf.add_tag('Sf_category', sf_category)
                    sf.add_tag('Sf_framecode', new_block_name)
                    if sf_category != 'entry_information':
                        sf.add_tag('Entry_ID', _entry_id)
                        sf.add_tag('ID', _cur_list_id)
                    else:
                        sf.add_tag('ID', _entry_id)

                    for _item, value in zip(itVals['Items'], itVals['Values'][0]):
                        if _item not in ('Sf_category', 'Sf_framecode', 'Entry_ID', 'ID'):
                            sf.add_tag(_item, value)

                if not item['sf_category_flag']:
                    lp = pynmrstar.Loop.from_scratch(category)

                    dBlockStruct = cifObj.getDataBlockStructure(block_name, ext)
                    itVals = next(v for k, v in dBlockStruct.items() if k == category)
                    lenItems = len(itVals['Items'])

                    list_id_idx = entry_id_idx = -1

                    for idx, _item in enumerate(itVals['Items']):
                        lp.add_tag(_item)
                        if sf_category != item['super_category']:
                            tag = '_' + category + '.' + _item
                            tag = tag.lower()
                            if tag in self.schema:
                                sdict = {k: v for k, v in self.schema[tag].items() if v not in emptyValue}
                                if 'Parent tag' in sdict and sdict['Parent tag'] == '_' + sf_tag_prefix + '.ID':
                                    list_id_idx = idx
                        if _item == 'Entry_ID':
                            entry_id_idx = idx

                    for idx, _value in enumerate(itVals['Values']):
                        if len(_value) != lenItems:  # prevent exception for incomplete stop tag at end of single loop
                            continue
                        if list_id_idx != -1:
                            _value[list_id_idx] = _cur_list_id
                        if entry_id_idx != -1:
                            _value[entry_id_idx] = _entry_id
                        lp.add_data(_value)

                    if sf is None:
                        sf = pynmrstar.Saveframe.from_scratch(block_name)
                        sf.set_tag_prefix(sf_tag_prefix)
                        sf.add_tag('Sf_category', sf_category)
                        sf.add_tag('Sf_framecode', block_name)
                        if sf_category != 'entry_information':
                            sf.add_tag('Entry_ID', _entry_id)
                            sf.add_tag('ID', _cur_list_id)
                        else:
                            sf.add_tag('ID', _entry_id)

                    sf.add_loop(lp)

                    if list_id_idx == -1:
                        set_lp_tag(lp, sf_tag_prefix + '_ID', _cur_list_id)

            if sf is not None:
                strData.add_saveframe(sf)

            retrieve_symbolic_labels(strData)

            self.normalize(strData)

            strData.write_to_file(strPath, show_comments=False, skip_empty_loops=True, skip_empty_tags=False)

            return True

        except ValueError as e:

            sf_anonymous_pattern = re.compile(r'\s*save_\S+\s*')
            save_pattern = re.compile(r'\s*save_\s*')

            split_ext = os.path.splitext(cifPath)
            _cifPath = split_ext[0] + '-corrected' + ('' if len(split_ext) == 1 else split_ext[1])

            changed = False

            with open(cifPath, 'r') as ifh, \
                    open(_cifPath, 'w') as ofh:
                for line in ifh:
                    if sf_anonymous_pattern.match(line) or save_pattern.match(line):
                        ofh.write('#' + line)
                        changed = True
                    else:
                        ofh.write(line)

            if changed and maxRepeat > 0:
                return self.convert(_cifPath, strPath, datablockName, originalFileName, maxRepeat - 1)

            try:
                os.remove(_cifPath)
            except OSError:
                pass

            self.__lfh.write(f"+{self.__class_name__}.convert() ++ Error  - {str(e)}\n")

            return False

        except Exception as e:
            self.__lfh.write(f"+{self.__class_name__}.convert() ++ Error  - {str(e)}\n")

            return False

    def set_entry_id(self, strData: Union[pynmrstar.Entry, pynmrstar.Saveframe, pynmrstar.Loop], entryId: str
                     ) -> bool:
        """ Set entry ID without changing datablock name.
            @see: pynmrstar.entry
            @return: whether document is modified
        """

        modified = False

        if strData is None:
            return modified

        try:
            sf = strData.frame_list[0]
            if sf.category.startswith('nef'):
                return False
        except (IndexError, AttributeError):
            pass

        entryId = entryId.strip().replace(' ', '_')  # DAOTHER-9511: replace white space in a datablock name to underscore

        if isinstance(strData, pynmrstar.Entry):

            for sf in strData.frame_list:
                filled = False

                for tag in sf.tags:
                    fqtn = (sf.tag_prefix + '.' + tag[0]).lower()

                    try:
                        if self.schema[fqtn]['entryIdFlg'] == 'Y':
                            tag[1] = entryId
                            filled = True
                            break
                    except KeyError:
                        pass

                if not filled:
                    entry_id_tag = 'ID' if sf.category == 'entry_information' else 'Entry_ID'

                    fqtn = (sf.tag_prefix + '.' + entry_id_tag).lower()

                    try:
                        if self.schema[fqtn]['entryIdFlg'] == 'Y':
                            set_sf_tag(sf, entry_id_tag, entryId)
                        modified = True
                    except KeyError:
                        pass

                for lp in sf.loops:
                    filled = False

                    for tag in lp.tags:
                        fqtn = (lp.category + '.' + tag).lower()

                        try:
                            if self.schema[fqtn]['entryIdFlg'] == 'Y' or tag == 'Entry_ID':
                                lp[tag] = [entryId] * len(lp[tag])
                                filled = True
                                break
                        except KeyError:
                            pass

                    if not filled:
                        set_lp_tag(lp, 'Entry_ID', entryId)

                        modified = True

        elif isinstance(strData, pynmrstar.Saveframe):

            sf = strData

            filled = False

            for tag in sf.tags:
                fqtn = (sf.tag_prefix + '.' + tag[0]).lower()

                try:
                    if self.schema[fqtn]['entryIdFlg'] == 'Y':
                        tag[1] = entryId
                        filled = True
                        break
                except KeyError:
                    pass

            if not filled:
                entry_id_tag = 'ID' if sf.category == 'entry_information' else 'Entry_ID'

                fqtn = (sf.tag_prefix + '.' + entry_id_tag).lower()

                try:
                    if self.schema[fqtn]['entryIdFlg'] == 'Y':
                        set_sf_tag(sf, entry_id_tag, entryId)
                        modified = True
                except KeyError:
                    pass

            for lp in sf.loops:
                filled = False

                for tag in lp.tags:
                    fqtn = (lp.category + '.' + tag).lower()

                    try:
                        if self.schema[fqtn]['entryIdFlg'] == 'Y' or tag == 'Entry_ID':
                            lp[tag] = [entryId] * len(lp[tag])
                            filled = True
                            break
                    except KeyError:
                        pass

                if not filled:
                    set_lp_tag(lp, 'Entry_ID', entryId)

                    modified = True

        elif isinstance(strData, pynmrstar.Loop):

            lp = strData

            filled = False

            for tag in lp.tags:
                fqtn = (lp.category + '.' + tag).lower()

                try:
                    if self.schema[fqtn]['entryIdFlg'] == 'Y' or tag == 'Entry_ID':
                        lp[tag] = [entryId] * len(lp[tag])
                        filled = True
                        break
                except KeyError:
                    pass

            if not filled:
                set_lp_tag(lp, 'Entry_ID', entryId)

                modified = True

        return modified

    def set_local_sf_id(self, strData: Union[pynmrstar.Entry, pynmrstar.Saveframe, pynmrstar.Loop], listId: Union[int, str]):
        """ Set list ID for a given saveframe or loop.
        """

        if strData is None:
            return

        try:
            sf = strData.frame_list[0]
            if sf.category.startswith('nef'):
                return
        except (IndexError, AttributeError):
            pass

        if isinstance(strData, pynmrstar.Saveframe):

            sf = strData

            for tag in sf.tags:
                fqtn = (sf.tag_prefix + '.' + tag[0]).lower()

                try:
                    if self.schema[fqtn]['lclSfIdFlg'] == 'Y':
                        tag[1] = listId
                        break
                except KeyError:
                    pass

            for lp in sf.loops:

                for tag in lp.tags:
                    fqtn = (lp.category + '.' + tag).lower()

                    try:
                        if self.schema[fqtn]['lclSfIdFlg'] == 'Y':
                            lp[tag] = [listId] * len(lp[tag])
                            break
                    except KeyError:
                        pass

        elif isinstance(strData, pynmrstar.Loop):

            lp = strData

            for tag in lp.tags:
                fqtn = (lp.category + '.' + tag).lower()

                try:
                    if self.schema[fqtn]['lclSfIdFlg'] == 'Y':
                        lp[tag] = [listId] * len(lp[tag])
                        break
                except KeyError:
                    pass

    def normalize(self, strData: Union[pynmrstar.Entry, pynmrstar.Saveframe, pynmrstar.Loop]
                  ) -> Union[pynmrstar.Entry, pynmrstar.Saveframe, pynmrstar.Loop]:
        """ Wrapper function of normalize_str() and normalize_nef().
        """

        if strData is None:
            return strData

        try:
            sf = strData.frame_list[0]
            if sf.category.startswith('nef'):
                return self.normalize_nef(strData)
            return self.normalize_str(strData)
        except (IndexError, AttributeError):
            return strData

    def normalize_str(self, strData: Union[pynmrstar.Entry, pynmrstar.Saveframe, pynmrstar.Loop]
                      ) -> Union[pynmrstar.Entry, pynmrstar.Saveframe, pynmrstar.Loop]:
        """ Sort saveframes, loops, and tags according to NMR-STAR schema.
            @see: pynmrstar.entry.normalize
        """

        if strData is None:
            return strData

        def sf_key(sf):
            """ Helper function to sort the saveframes.
                Returns (category order, saveframe order)
            """

            # If not a real category, generate an artificial but stable order > the real saveframes
            try:
                category_order = self.category_order.index(sf.tag_prefix)
            except (ValueError, KeyError):
                if sf.category is None:
                    category_order = float('infinity')
                else:
                    category_order = len(self.category_order) + abs(int(hashlib.sha1(str(sf.category).encode()).hexdigest(), 16))

            # See if there is an ID tag, and it is a number
            saveframe_id = float('infinity')
            try:
                saveframe_id = int(sf.get_tag("ID")[0])
            except (ValueError, KeyError, IndexError, TypeError):
                # Either there is no ID, or it is not a number. By default it will sort at the end of saveframes of its
                # category. Note that the entry_information ID tag has a different meaning, but since there should
                # only ever be one saveframe of that category, the sort order for it can be any value.
                pass

            return category_order, saveframe_id

        def lp_key(lp):
            """ Helper function to sort the loops.
                Returns (category order)
            """

            try:
                category_order = self.category_order.index(lp.category)
            except (ValueError, KeyError):
                category_order = float('infinity')

            return category_order

        try:
            strData.frame_list.sort(key=sf_key)

            for sf in strData.frame_list:
                sf.sort_tags()
                if len(sf.loops) > 1:
                    sf.loops.sort(key=lp_key)
                # Iterate through the loops
                for lp in sf:
                    lp.sort_tags()

        except Exception as e:
            self.__lfh.write(f"+{self.__class_name__}.normalize() ++ Error  - {str(e)}\n")

        return strData

    def normalize_nef(self, strData: Union[pynmrstar.Entry, pynmrstar.Saveframe, pynmrstar.Loop]
                      ) -> Union[pynmrstar.Entry, pynmrstar.Saveframe, pynmrstar.Loop]:
        """ Sort saveframes of NEF.
        """

        if strData is None:
            return strData

        def sf_key(sf):
            """ Helper function to sort the saveframes.
                Returns (saveframe order)
            """

            try:
                category_order = self.category_order_nef.index(sf.tag_prefix)
            except (ValueError, KeyError):
                if sf.category is None:
                    category_order = float('infinity')
                else:
                    category_order = len(self.category_order_nef) + abs(int(hashlib.sha1(str(sf.category).encode()).hexdigest(), 16))

            return category_order

        try:
            strData.frame_list.sort(key=sf_key)
        except Exception as e:
            self.__lfh.write(f"+{self.__class_name__}.normalize_nef() ++ Error  - {str(e)}\n")

        return strData

    def cleanup(self, strData: pynmrstar.Entry) -> pynmrstar.Entry:
        """ Wrapper function of cleanup_str() and cleanup_nef().
        """

        if strData is None:
            return strData

        try:
            sf = strData.frame_list[0]
            if sf.category.startswith('nef'):
                return self.cleanup_nef(strData)
            return self.cleanup_str(strData)
        except (IndexError, AttributeError):
            return strData

    def cleanup_str(self, strData: pynmrstar.Entry) -> pynmrstar.Entry:
        """ Remove empty/ineffective saveframes and loops of NMR-STAR.
        """

        strData.remove_empty_saveframes()

        list_id_dict = {}
        empty_saveframes = []

        for sf in strData.frame_list:

            if sf.category not in list_id_dict:
                list_id_dict[sf.category] = 1
            else:
                list_id_dict[sf.category] += 1

            has_list_id = any(tag[0] == 'ID' for tag in sf.tags)

            if has_list_id and sf.category != 'entry_information':
                list_id_tag = 'ID' if sf.category != 'chem_comp' else 'PDB_code'
                list_id = next((tag[1] for tag in sf.tags if tag[0] == list_id_tag), None)
                if list_id not in emptyValue:
                    self.set_local_sf_id(sf, list_id_dict[sf.category]
                                         if isinstance(list_id, int) or list_id.isdigit() else list_id)

            has_eff_sf_tag = any(tag[1] not in emptyValue for tag in sf.tags
                                 if tag[0] not in ('Sf_category', 'Sf_framecode', 'Entry_ID', 'Sf_ID', 'ID'))

            if sf.category == 'NMR_spectrometer_expt':
                if get_first_sf_tag(sf, 'Name') in emptyValue:
                    has_eff_sf_tag = False

            if len(get_first_sf_tag(sf, 'Sf_framecode')) == 0:
                has_eff_sf_tag = False

            entry_info_sf = sf.category == 'entry_information'

            list_id_tag = alt_list_id_tag = None
            if not entry_info_sf:
                list_id_tag = f'{sf.tag_prefix[1:]}_ID'
                if sf.tag_prefix.endswith('_list'):
                    alt_list_id_tag = f'{sf.tag_prefix[1:-5]}_ID'

            empty_loops = []

            for lp in sf.loops:

                if entry_info_sf and lp.category == '_Release':
                    continue

                if lp.category.startswith('_PDBX_'):
                    empty_loops.append(lp)
                    continue

                entry_id_col = lp.tags.index('Entry_ID') if 'Entry_ID' in lp.tags else -1
                sf_id_col = lp.tags.index('Sf_ID') if 'Sf_ID' in lp.tags else -1
                if entry_info_sf:
                    list_id_col = -1
                else:
                    list_id_col = lp.tags.index(list_id_tag) if list_id_tag in lp.tags else -1
                    if list_id_col == -1 and alt_list_id_tag is not None:
                        list_id_col = lp.tags.index(alt_list_id_tag) if alt_list_id_tag in lp.tags else -1

                id_tag = None
                if 'ID' in lp.tags:
                    id_tag = 'ID'
                elif 'Ordinal' in lp.tags:
                    id_tag = 'Ordinal'
                else:
                    id_tags = [tag for tag in lp.tags
                               if tag.endswith('ID') and tag not in ('Sf_ID', 'Entry_ID', list_id_tag, alt_list_id_tag)]
                    if len(id_tags) == 1:
                        id_tag = id_tags[0]

                id_col = lp.tags.index(id_tag) if id_tag is not None else -1

                empty_row_idx = []

                for idx, row in enumerate(lp):

                    if 'Name' in lp.tags:
                        if row[lp.tags.index('Name')] in emptyValue:
                            if lp.category != '_NMR_spectrometer_view':
                                empty_row_idx.append(idx)
                                continue

                            try:
                                vendor = row[lp.tags.index('Manufacturer')]
                                model = row[lp.tags.index('Model')]
                                field = row[lp.tags.index('Field_strength')]
                                if vendor in emptyValue\
                                   or model in emptyValue\
                                   or field in emptyValue:
                                    empty_row_idx.append(idx)
                                    continue
                            except ValueError:
                                empty_row_idx.append(idx)
                                continue

                            for parent_sf in strData.get_saveframes_by_tag_and_value('_NMR_spectrometer.Manufacturer', vendor):
                                if get_first_sf_tag(parent_sf, 'Model') == model\
                                   and get_first_sf_tag(parent_sf, 'Field_strength', field) == field:
                                    row[lp.tags.index('Name')] = get_first_sf_tag(parent_sf, 'Sf_framecode')
                                    break

                    if lp.category == '_Assembly_db_link':
                        if 'Accession_code' in lp.tags and row[lp.tags.index('Accession_code')] in emptyValue:
                            empty_row_idx.append(idx)
                            continue

                    if lp.category == '_Entity_purity':
                        if 'Val' in lp.tags and row[lp.tags.index('Val')] in emptyValue:
                            empty_row_idx.append(idx)
                            continue

                    if any(row[col] not in emptyValue for col in range(len(row))
                           if col not in (entry_id_col, sf_id_col, list_id_col, id_col)):
                        continue

                    empty_row_idx.append(idx)

                if len(empty_row_idx) > 0:
                    for idx in reversed(empty_row_idx):
                        del lp.data[idx]

                if len(lp) == 0:
                    empty_loops.append(lp)
                    continue

                if id_col != -1:
                    for idx, row in enumerate(lp, start=1):
                        if row[id_col] in emptyValue:
                            row[id_col] = idx

            if len(empty_loops) > 0:
                for lp in empty_loops:
                    del sf[lp]

            if not has_eff_sf_tag and len(sf.loops) == 0:
                empty_saveframes.append(sf)

        if len(empty_saveframes) > 0:
            for sf in empty_saveframes:
                strData.remove_saveframe(sf)

        return strData

    def cleanup_nef(self, strData: pynmrstar.Entry) -> pynmrstar.Entry:  # pylint: disable=no-self-use
        """ Remove empty/ineffective saveframes and loops of NEF.
        """

        strData.remove_empty_saveframes()

        empty_saveframes = []

        for sf in strData.frame_list:

            has_eff_sf_tag = any(tag[1] not in emptyValue for tag in sf.tags
                                 if tag[0] not in ('sf_category', 'sf_framecode'))

            if len(get_first_sf_tag(sf, 'sf_framecode')) == 0:
                has_eff_sf_tag = False

            empty_loops = []

            for lp in sf.loops:

                id_tag = None
                id_tags = [tag for tag in lp.tags if tag.endswith('_id')]
                if len(id_tags) == 1:
                    id_tag = id_tags[0]

                id_col = lp.tags.index(id_tag) if id_tag is not None else -1

                empty_row_idx = []

                for idx, row in enumerate(lp):

                    if any(row[col] not in emptyValue for col in range(len(row)) if col != id_col):
                        continue

                    empty_row_idx.append(idx)

                if len(empty_row_idx) > 0:
                    for idx in reversed(empty_row_idx):
                        del lp.data[idx]

                if len(lp) == 0:
                    empty_loops.append(lp)
                    continue

                if id_col != -1:
                    for idx, row in enumerate(lp, start=1):
                        if row[id_col] in emptyValue:
                            row[id_col] = idx

            if len(empty_loops) > 0:
                for lp in empty_loops:
                    del sf[lp]

            if not has_eff_sf_tag and len(sf.loops) == 0:
                empty_saveframes.append(sf)

        if len(empty_saveframes) > 0:
            for sf in empty_saveframes:
                strData.remove_saveframe(sf)

        return strData
